def extract_doxygen_method_params(doxystr):
    """
    Given a doxygen string for a method, extract parameter descriptions
    """
    doxyVarDesc = {}
    doxyLines = doxystr.split("\n")
    lastParamDesc = ""
    for doxyLine in doxyLines:
        if " @param " in doxyLine or " \\param " in doxyLine:
            try:
                # Strip out the param
                doxyLine = doxyLine[doxyLine.find("param ") + 6 :]
                (var, desc) = doxyLine.split(" ", 1)
                doxyVarDesc[var] = desc.strip()
                lastParamDesc = var
            except:
                pass
        elif " @return " in doxyLine or " \return " in doxyLine:
            lastParamDesc = ""
            # not handled for now
        elif lastParamDesc:
            try:
                doxyLine = doxyLine.strip()
                if " " not in doxyLine:
                    lastParamDesc = ""
                    continue
                doxyLine = doxyLine[doxyLine.find(" ") + 1 :]
                doxyVarDesc[lastParamDesc] += " " + doxyLine
            except:
                pass

    return doxyVarDesc



from collections import deque
import ply.lex as lex
import re

_line_re = re.compile(r'^#line (\d+) "(.*)"')


class Lexer(object):

    tokens = [
        "NUMBER",
        "FLOAT_NUMBER",
        "NAME",
        "COMMENT_SINGLELINE",
        "COMMENT_MULTILINE",
        "PRECOMP_MACRO",
        "PRECOMP_MACRO_CONT",
        "DIVIDE",
        "CHAR_LITERAL",
        "STRING_LITERAL",
        "NEWLINE",
        "ELLIPSIS",
        "DBL_LBRACKET",
        "DBL_RBRACKET",
        "DBL_COLON",
        "SHIFT_LEFT",
    ]

    literals = [
        "<",
        ">",
        "(",
        ")",
        "{",
        "}",
        "[",
        "]",
        ";",
        ":",
        ",",
        "\\",
        "|",
        "%",
        "^",
        "!",
        "*",
        "-",
        "+",
        "&",
        "=",
        "'",
        ".",
    ]

    t_ignore = " \t\r?@\f"
    t_NUMBER = r"[0-9][0-9XxA-Fa-f]*"
    t_FLOAT_NUMBER = r"[-+]?[0-9]*\.[0-9]+([eE][-+]?[0-9]+)?"
    t_NAME = r"[A-Za-z_~][A-Za-z0-9_]*"

    def t_PRECOMP_MACRO(self, t):
        r"\#.*"
        m = _line_re.match(t.value)
        if m:
            filename = m.group(2)
            if filename not in self._filenames_set:
                self.filenames.append(filename)
                self._filenames_set.add(filename)
            self.filename = filename

            self.line_offset = 1 + self.lex.lineno - int(m.group(1))

        else:
            return t

    t_PRECOMP_MACRO_CONT = r".*\\\n"

    def t_COMMENT_SINGLELINE(self, t):
        r"\/\/.*\n?"
        if t.value.startswith("///") or t.value.startswith("//!"):
            self.comments.append(t.value.lstrip("\t ").rstrip("\n"))
        t.lexer.lineno += t.value.count("\n")
        return t

    t_DIVIDE = r"/(?!/)"
    t_CHAR_LITERAL = "'.'"
    t_ELLIPSIS = r"\.\.\."
    t_DBL_LBRACKET = r"\[\["
    t_DBL_RBRACKET = r"\]\]"
    t_DBL_COLON = r"::"
    t_SHIFT_LEFT = r"<<"
    # SHIFT_RIGHT introduces ambiguity

    # found at http://wordaligned.org/articles/string-literals-and-regular-expressions
    # TODO: This does not work with the string "bla \" bla"
    t_STRING_LITERAL = r'"([^"\\]|\\.)*"'

    # Found at http://ostermiller.org/findcomment.html
    def t_COMMENT_MULTILINE(self, t):
        r"/\*([^*]|[\r\n]|(\*+([^*/]|[\r\n])))*\*+/\n?"
        if t.value.startswith("/**") or t.value.startswith("/*!"):
            # not sure why, but get double new lines
            v = t.value.replace("\n\n", "\n")
            # strip prefixing whitespace
            v = re.sub("\n[\\s]+\\*", "\n*", v)
            self.comments = v.splitlines()
        t.lexer.lineno += t.value.count("\n")
        return t

    def t_NEWLINE(self, t):
        r"\n+"
        t.lexer.lineno += len(t.value)
        del self.comments[:]
        return t

    def t_error(self, v):
        print("Lex error: ", v)

    def __init__(self, filename):
        self.lex = lex.lex(module=self)
        self.input = self.lex.input

        # For tracking current file/line position
        self.filename = filename
        self.line_offset = 0

        self.filenames = []
        self._filenames_set = set()

        if self.filename:
            self.filenames.append(filename)
            self._filenames_set.add(filename)

        # Doxygen comments
        self.comments = []

        self.lookahead = deque()

    def current_location(self):
        if self.lookahead:
            return self.lookahead[0].location
        return self.filename, self.lex.lineno - self.line_offset

    def get_doxygen(self):
        """
        This should be called after the first element of something has
        been consumed.

        It will lookahead for comments that come after the item, if prior
        comments don't exist.
        """

        # Assumption: This function is either called at the beginning of a
        # statement or at the end of a statement

        if self.comments:
            comments = self.comments
        else:
            comments = []
            # only look for comments until a newline (including lookahead)
            for tok in self.lookahead:
                if tok.type == "NEWLINE":
                    return ""

            while True:
                tok = self.lex.token()
                comments.extend(self.comments)

                if tok is None:
                    break

                tok.location = (self.filename, tok.lineno - self.line_offset)
                ttype = tok.type
                if ttype == "NEWLINE":
                    self.lookahead.append(tok)
                    break

                if ttype not in self._discard_types:
                    self.lookahead.append(tok)

                if ttype == "NAME":
                    break

                del self.comments[:]

        comments = "\n".join(comments)
        del self.comments[:]
        return comments

    _discard_types = {"NEWLINE", "COMMENT_SINGLELINE", "COMMENT_MULTILINE"}

    def token(self, eof_ok=False):
        tok = None
        while self.lookahead:
            tok = self.lookahead.popleft()
            if tok.type not in self._discard_types:
                return tok

        while True:
            tok = self.lex.token()
            if tok is None:
                if not eof_ok:
                    raise EOFError("unexpected end of file")
                break

            if tok.type not in self._discard_types:
                tok.location = (self.filename, tok.lineno - self.line_offset)
                break

        return tok

    def token_if(self, *types):
        tok = self.token(eof_ok=True)
        if tok is None:
            return None
        if tok.type not in types:
            # put it back on the left in case it was retrieved
            # from the lookahead buffer
            self.lookahead.appendleft(tok)
            return None
        return tok

    def return_token(self, tok):
        self.lookahead.appendleft(tok)

    def return_tokens(self, toks):
        self.lookahead.extendleft(reversed(toks))


if __name__ == "__main__":
    try:
        lex.runmain(lexer=Lexer(None))
    except EOFError:
        pass



#!/usr/bin/python3 -i
#
# Copyright (c) 2018 Valve Corporation
# Copyright (c) 2018 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from base_generator import BaseGenerator, BaseGeneratorOptions, write
from base_struct_decoders_header_generator import BaseStructDecodersHeaderGenerator


class VulkanStructDecodersHeaderGeneratorOptions(BaseGeneratorOptions):
    """Options for generating C++ type declarations for Vulkan struct decoding."""

    def __init__(
        self,
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extraVulkanHeaders=[]
    ):
        BaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extraVulkanHeaders=extraVulkanHeaders
        )


class VulkanStructDecodersHeaderGenerator(
    BaseStructDecodersHeaderGenerator, BaseGenerator
):
    """VulkanStructDecodersHeaderGenerator - subclass of BaseGenerator.
    Generates C++ type declarations for the decoded Vulkan API structure wrappers.
    Generate C++ type declarations for Vulkan struct decoding.
    """

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        BaseGenerator.__init__(
            self,
            process_cmds=False,
            process_structs=True,
            feature_break=True,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

    def beginFile(self, gen_opts):
        """Method override."""
        BaseGenerator.beginFile(self, gen_opts)

        write(
            '#include "decode/custom_vulkan_struct_decoders_forward.h"',
            file=self.outFile
        )
        write('#include "decode/handle_pointer_decoder.h"', file=self.outFile)
        write('#include "decode/pointer_decoder.h"', file=self.outFile)
        write('#include "decode/string_array_decoder.h"', file=self.outFile)
        write('#include "decode/string_decoder.h"', file=self.outFile)
        write('#include "decode/struct_pointer_decoder.h"', file=self.outFile)
        write('#include "decode/vulkan_pnext_node.h"', file=self.outFile)
        write('#include "format/format.h"', file=self.outFile)
        write('#include "format/platform_types.h"', file=self.outFile)
        write(
            '#include "generated/generated_vulkan_struct_decoders_forward.h"',
            file=self.outFile
        )
        write('#include "util/defines.h"', file=self.outFile)
        self.newline()
        self.includeVulkanHeaders(gen_opts)
        self.newline()
        write('#include <memory>', file=self.outFile)
        self.newline()
        write('GFXRECON_BEGIN_NAMESPACE(gfxrecon)', file=self.outFile)
        write('GFXRECON_BEGIN_NAMESPACE(decode)', file=self.outFile)

    def endFile(self):
        """Method override."""
        self.newline()
        write('GFXRECON_END_NAMESPACE(decode)', file=self.outFile)
        write('GFXRECON_END_NAMESPACE(gfxrecon)', file=self.outFile)

        # Finish processing in superclass
        BaseGenerator.endFile(self)

    def need_feature_generation(self):
        """Indicates that the current feature has C++ code to generate."""
        if self.feature_struct_members or self.feature_struct_aliases:
            return True
        return False



#!/usr/bin/env python3
#
# Copyright (c) 2018-2023 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import argparse
import os
import sys
import shlex
import subprocess

argv = sys.argv
argc = len(sys.argv)

# Supported commands
valid_commands = [
    'install-apk',
    'replay'
]

# Arguments
# gfxrecon install-apk <file>
# gfxrecon replay [-p | --push-file <file-on-desktop>] <file-on-device>

# Application info
app_name = 'com.lunarg.gfxreconstruct.replay'
app_activity = '"com.lunarg.gfxreconstruct.replay/android.app.NativeActivity"'
app_action = 'android.intent.action.MAIN'
app_category = 'android.intent.category.LAUNCHER'

# ADB commands
adb_install = 'adb install -g -t -r'
adb_sdk_version = 'adb shell getprop ro.build.version.sdk'
adb_start = 'adb shell am start -n {} -a {} -c {}'.format(app_activity, app_action, app_category)
adb_stop = 'adb shell am force-stop {}'.format(app_name)
adb_push = 'adb push'

def CreateCommandParser():
    parser = argparse.ArgumentParser(description='GFXReconstruct utility launcher for Android.')
    parser.add_argument('command', choices=valid_commands, metavar='command', help='Command to execute. Valid options are [{}]'.format(', '.join(valid_commands)))
    parser.add_argument('args', nargs=argparse.REMAINDER, help='Command-specific argument list. Specify -h after command name for more.')
    return parser

def CreateInstallApkParser():
    parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]) + ' install-apk', description='Install the replay tool.')
    parser.add_argument('file', help='APK file to install')
    return parser

def CreateReplayParser():
    parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]) + ' replay', description='Launch the replay tool.')
    parser.add_argument('-p', '--push-file', metavar='LOCAL_FILE', help='Local file to push to the location on device specified by <file>')
    parser.add_argument('--version', action='store_true', default=False, help='Print version information and exit (forwarded to replay tool)')
    parser.add_argument('--log-level', metavar='LEVEL', help='Specify highest level message to log. Options are: debug, info, warning, error, and fatal. Default is info. (forwarded to replay tool)')
    parser.add_argument('--log-file', metavar='DEVICE_FILE', help='Write log messages to a file at the specified path instead of logcat (forwarded to replay tool)')
    parser.add_argument('--pause-frame', metavar='N', help='Pause after replaying frame number N (forwarded to replay tool)')
    parser.add_argument('--paused', action='store_true', default=False, help='Pause after replaying the first frame (same as "--pause-frame 1"; forwarded to replay tool)')
    parser.add_argument('--screenshot-all', action='store_true', default=False, help='Generate screenshots for all frames.  When this option is specified, --screenshots is ignored (forwarded to replay tool)')
    parser.add_argument('--screenshots', metavar='RANGES', help='Generate screenshots for the specified frames.  Target frames are specified as a comma separated list of frame ranges.  A frame range can be specified as a single value, to specify a single frame, or as two hyphenated values, to specify the first and last frames to process.  Frame ranges should be specified in ascending order and cannot overlap.  Note that frame numbering is 1-based (i.e. the first frame is frame 1).  Example: 200,301-305 will generate six screenshots (forwarded to replay tool)')
    parser.add_argument('--screenshot-format', metavar='FORMAT', choices=['bmp', 'png'], help='Image file format to use for screenshot generation.  Available formats are: bmp, png (forwarded to replay tool)')
    parser.add_argument('--screenshot-dir', metavar='DIR', help='Directory to write screenshots. Default is "/sdcard" (forwarded to replay tool)')
    parser.add_argument('--screenshot-prefix', metavar='PREFIX', help='Prefix to apply to the screenshot file name.  Default is "screenshot" (forwarded to replay tool)')
    parser.add_argument('--screenshot-size', metavar='SIZE', help='Screenshot dimensions. Ignored if --screenshot-scale is specified.  Expected format is <width>x<height>.')
    parser.add_argument('--screenshot-scale', metavar='SCALE', help='Scale screenshot dimensions. Overrides --screenshot-size, if specified. Expects a number which can be decimal')
    parser.add_argument('--sfa', '--skip-failed-allocations', action='store_true', default=False, help='Skip vkAllocateMemory, vkAllocateCommandBuffers, and vkAllocateDescriptorSets calls that failed during capture (forwarded to replay tool)')
    parser.add_argument('--opcd', '--omit-pipeline-cache-data', action='store_true', default=False, help='Omit pipeline cache data from calls to vkCreatePipelineCache and skip calls to vkGetPipelineCacheData (forwarded to replay tool)')
    parser.add_argument('--surface-index', metavar='N', help='Restrict rendering to the Nth surface object created.  Used with captures that include multiple surfaces.  Default is -1 (render to all surfaces; forwarded to replay tool)')
    parser.add_argument('--sync', action='store_true', default=False, help='Synchronize after each queue submission with vkQueueWaitIdle (forwarded to replay tool)')
    parser.add_argument('--remove-unsupported', action='store_true', default=False, help='Remove unsupported extensions and features from instance and device creation parameters (forwarded to replay tool)')
    parser.add_argument('--validate', action='store_true', default=False, help='Enables the Khronos Vulkan validation layer (forwarded to replay tool)')
    parser.add_argument('--onhb', '--omit-null-hardware-buffers', action='store_true', default=False, help='Omit Vulkan calls that would pass a NULL AHardwareBuffer* (forwarded to replay tool)')
    parser.add_argument('--use-colorspace-fallback', action='store_true', default=False, help='Swap the swapchain color space if unsupported by replay device. Check if color space is not supported by replay device and swap to VK_COLOR_SPACE_SRGB_NONLINEAR_KHR. (forwarded to replay tool).')
    parser.add_argument('--offscreen-swapchain-frame-boundary', action='store_true', default=False, help='Should only be used with offscreen swapchain. Activates the extension VK_EXT_frame_boundary (always supported if trimming, checks for driver support otherwise) and inserts command buffer submission with VkFrameBoundaryEXT where vkQueuePresentKHR was called in the original capture. This allows preserving frames when capturing a replay that uses. offscreen swapchain. (forwarded to replay tool)')
    parser.add_argument('--mfr', '--measurement-frame-range', metavar='START-END', help='Custom framerange to measure FPS for. This range will include the start frame but not the end frame. The measurement frame range defaults to all frames except the loading frame but can be configured for any range. If the end frame is past the last frame in the trace it will be clamped to the frame after the last (so in that case the results would include the last frame). (forwarded to replay tool)')
    parser.add_argument('--measurement-file', metavar='DEVICE_FILE', help='Write measurements to a file at the specified path. Default is: \'/sdcard/gfxrecon-measurements.json\' on android and \'./gfxrecon-measurements.json\' on desktop. (forwarded to replay tool)')
    parser.add_argument('--quit-after-measurement-range', action='store_true', default=False, help='If this is specified the replayer will abort when it reaches the <end_frame> specified in the --measurement-frame-range argument. (forwarded to replay tool)')
    parser.add_argument('--flush-measurement-range', action='store_true', default=False, help='If this is specified the replayer will flush and wait for all current GPU work to finish at the start and end of the measurement range. (forwarded to replay tool)')
    parser.add_argument('--flush-inside-measurement-range', action='store_true', default=False, help='If this is specified the replayer will flush and wait for all current GPU work to finish at end of each frame inside the measurement range. (forwarded to replay tool)')
    parser.add_argument('--sgfs', '--skip-get-fence-status', metavar='STATUS', default=0, help='Specify behaviour to skip calls to vkWaitForFences and vkGetFenceStatus. Default is 0 - No skip (forwarded to replay tool)')
    parser.add_argument('--sgfr', '--skip-get-fence-ranges', metavar='FRAME-RANGES', default='', help='Frame ranges where --sgfs applies. Default is all frames (forwarded to replay tool)')
    parser.add_argument('--wait-before-present', action='store_true', default=False, help='Force wait on completion of queue operations for all queues before calling Present. This is needed for accurate acquisition of instrumentation data on some platforms.')
    parser.add_argument('-m', '--memory-translation', metavar='MODE', choices=['none', 'remap', 'realign', 'rebind'], help='Enable memory translation for replay on GPUs with memory types that are not compatible with the capture GPU\'s memory types.  Available modes are: none, remap, realign, rebind (forwarded to replay tool)')
    parser.add_argument('--swapchain', metavar='MODE', choices=['virtual', 'captured', 'offscreen'], help='Choose a swapchain mode to replay. Available modes are: virtual, captured, offscreen (forwarded to replay tool)')
    parser.add_argument('--vssb', '--virtual-swapchain-skip-blit', action='store_true', default=False, help='Skip blit to real swapchain to gain performance during replay.')
    parser.add_argument('--use-captured-swapchain-indices', action='store_true', default=False, help='Same as "--swapchain captured". Ignored if the "--swapchain" option is used.')
    parser.add_argument('file', nargs='?', help='File on device to play (forwarded to replay tool)')
    return parser

def MakeExtrasString(args):
    arg_list = []

    if args.version:
        arg_list.append('--version')

    if args.log_level:
        arg_list.append('--log-level')
        arg_list.append('{}'.format(args.log_level))

    if args.log_file:
        arg_list.append('--log-file')
        arg_list.append('{}'.format(args.log_file))

    if args.pause_frame:
        arg_list.append('--pause-frame')
        arg_list.append('{}'.format(args.pause_frame))

    if args.paused:
        arg_list.append('--paused')

    if args.screenshot_all:
        arg_list.append('--screenshot-all')
    elif args.screenshots:
        arg_list.append('--screenshots')
        arg_list.append('{}'.format(args.screenshots))

    if args.screenshot_format:
        arg_list.append('--screenshot-format')
        arg_list.append('{}'.format(args.screenshot_format))

    if args.screenshot_dir:
        arg_list.append('--screenshot-dir')
        arg_list.append('{}'.format(args.screenshot_dir))

    if args.screenshot_prefix:
        arg_list.append('--screenshot-prefix')
        arg_list.append('{}'.format(args.screenshot_prefix))

    if args.screenshot_size:
        arg_list.append('--screenshot-size')
        arg_list.append('{}'.format(args.screenshot_size))

    if args.screenshot_scale:
        arg_list.append('--screenshot-scale')
        arg_list.append('{}'.format(args.screenshot_scale))

    if args.sfa:
        arg_list.append('--sfa')

    if args.opcd:
        arg_list.append('--opcd')

    if args.surface_index:
        arg_list.append('--surface-index')
        arg_list.append('{}'.format(args.surface_index))

    if args.sync:
        arg_list.append('--sync')

    if args.remove_unsupported:
        arg_list.append('--remove-unsupported')

    if args.validate:
        arg_list.append('--validate')

    if args.onhb:
        arg_list.append('--onhb')

    if args.use_captured_swapchain_indices:
        arg_list.append('--use-captured-swapchain-indices')

    if args.use_colorspace_fallback:
        arg_list.append('--use-colorspace-fallback')

    if args.mfr:
        arg_list.append('--mfr')
        arg_list.append('{}'.format(args.mfr))

    if args.measurement_file:
        arg_list.append('--measurement-file')
        arg_list.append('{}'.format(args.measurement_file))

    if args.quit_after_measurement_range:
        arg_list.append('--quit-after-measurement-range')

    if args.flush_measurement_range:
        arg_list.append('--flush-measurement-range')

    if args.flush_inside_measurement_range:
        arg_list.append('--flush-inside-measurement-range')

    if args.swapchain:
        arg_list.append('--swapchain')
        arg_list.append('{}'.format(args.swapchain))

    if args.offscreen_swapchain_frame_boundary:
        arg_list.append('--offscreen-swapchain-frame-boundary')

    if args.vssb:
        arg_list.append('--vssb')

    if args.sgfs:
        arg_list.append('--sgfs')
        arg_list.append('{}'.format(args.sgfs))

    if args.sgfr:
        arg_list.append('--sgfr')
        arg_list.append('{}'.format(args.sgfr))

    if args.memory_translation:
        arg_list.append('-m')
        arg_list.append('{}'.format(args.memory_translation))

    if args.wait_before_present:
        arg_list.append('--wait-before-present')

    if args.file:
        arg_list.append(args.file)
    elif not args.version:
        print('gfxrecon.py release: error: the following arguments are required: file')
        return None

    return ' '.join(arg_list)

def InstallApk(install_args):
    install_parser = CreateInstallApkParser()
    args = install_parser.parse_args(install_args)
    sdk = int(subprocess.check_output(shlex.split(adb_sdk_version)).decode())
    force_queryable = ' --force-queryable' if sdk >= 30 else ''
    cmd = adb_install + force_queryable + ' ' + args.file
    print('Executing:', cmd)
    subprocess.check_call(shlex.split(cmd, posix='win' not in sys.platform))

def Replay(replay_args):
    replay_parser = CreateReplayParser()
    args = replay_parser.parse_args(replay_args)

    extras = MakeExtrasString(args)

    if extras:
        if args.push_file:
            cmd = ' '.join([adb_push, args.push_file, args.file])
            print('Executing:', cmd)
            subprocess.check_call(shlex.split(cmd, posix='win' not in sys.platform))

        print('Executing:', adb_stop)
        subprocess.check_call(shlex.split(adb_stop, posix='win' not in sys.platform))

        cmd = ' '.join([adb_start, '--es', '"args"', '"{}"'.format(extras)])
        print('Executing:', cmd)

        # Specify posix=False to prevent removal of quotes from adb extras.
        subprocess.check_call(shlex.split(cmd, posix=False))

if __name__ == '__main__':
    command_parser = CreateCommandParser()
    command = command_parser.parse_args()

    if command.command == 'install-apk':
        InstallApk(command.args)
    elif command.command == 'replay':
        Replay(command.args)



#!/usr/bin/python3 -i
#
# Copyright (c) 2022-2023 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#

import sys
from base_generator import *
from reformat_code import format_cpp_code, indent_cpp_code, remove_trailing_newlines


class VulkanStructToJsonHeaderGeneratorOptions(BaseGeneratorOptions):
    """Options for generating C++ functions for to_json functions"""

    def __init__(
        self,
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extraVulkanHeaders=[]
    ):
        BaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extraVulkanHeaders=extraVulkanHeaders
        )


# VulkanStructToJsonHeaderGenerator - subclass of BaseGenerator.
# Generates C++ functions for stringifying Vulkan API structures.
class VulkanStructToJsonHeaderGenerator(BaseGenerator):
    """Generate C++ functions to serialize Vulkan structures to JSON"""

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        BaseGenerator.__init__(
            self,
            process_cmds=False,
            process_structs=True,
            feature_break=True,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

        self.customImplementationRequired = {
            'VkPipelineCacheCreateInfo',
            'VkShaderModuleCreateInfo',
            'VkPipelineExecutableStatisticKHR',
        }

    # Method override
    # yapf: disable
    def beginFile(self, genOpts):
        BaseGenerator.beginFile(self, genOpts)
        body = format_cpp_code('''
            #include "decode/custom_vulkan_struct_to_json.h"

            GFXRECON_BEGIN_NAMESPACE(gfxrecon)
            GFXRECON_BEGIN_NAMESPACE(decode)
            ''')
        body += "\n"
        write(body, file=self.outFile)
    # yapf: enable

    # Method override
    # yapf: disable
    def endFile(self):
        body = remove_trailing_newlines(indent_cpp_code('''
            /// Works out the type of the struct at the end of a pNext pointer and dispatches
            /// recursively to the FieldToJson for that.
            void FieldToJson(nlohmann::ordered_json& jdata, const PNextNode* data, const util::JsonOptions& options = util::JsonOptions());

            GFXRECON_END_NAMESPACE(decode)
            GFXRECON_END_NAMESPACE(gfxrecon)
        '''))
        write(body, file=self.outFile)

        # Finish processing in superclass
        BaseGenerator.endFile(self)
    # yapf: enable

    #
    # Indicates that the current feature has C++ code to generate.
    def need_feature_generation(self):
        self.feature_break = False
        if self.feature_struct_members:
            return True
        return False

    #
    # Performs C++ code generation for the feature.
    # yapf: disable
    def generate_feature(self):
        for struct in self.get_filtered_struct_names():
            if not struct in self.customImplementationRequired:
                body = "void FieldToJson(nlohmann::ordered_json& jdata, const Decoded_{0}* data, const util::JsonOptions& options = util::JsonOptions());".format(struct)
                write(body, file=self.outFile)
    # yapf: enable




#!/usr/bin/python3 -i
#
# Copyright (c) 2023 LunarG, Inc.
# Copyright (c) 2023 Arm Limited and/or its affiliates <open-source-office@arm.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from base_generator import BaseGenerator, BaseGeneratorOptions, write


class VulkanStructTrackersBodyGeneratorOptions(BaseGeneratorOptions):
    """Options for generating function definitions to track (deepcopy) Vulkan structs at API capture for trimming."""

    def __init__(
        self,
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extraVulkanHeaders=[]
    ):
        BaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extraVulkanHeaders=extraVulkanHeaders
        )


class VulkanStructTrackersBodyGenerator(BaseGenerator):
    """VulkanStructTrackersHeaderGenerator - subclass of BaseGenerator.
    Generates C++ function definitions to track (deepcopy) Vulkan structs
    at API capture for trimming.
    """

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        BaseGenerator.__init__(
            self,
            process_cmds=True,
            process_structs=True,
            feature_break=False,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

        # Map of typename to VkStructureType for each struct that is not an alias and has a VkStructureType associated
        self.struct_type_enums = dict()

    def beginFile(self, gen_opts):
        """Method override."""
        BaseGenerator.beginFile(self, gen_opts)

        write(
            '#include "generated/generated_vulkan_struct_handle_wrappers.h"',
            file=self.outFile
        )
        write('#include "generated/generated_vulkan_struct_trackers.h"', file=self.outFile)
        self.newline()
        write('GFXRECON_BEGIN_NAMESPACE(gfxrecon)', file=self.outFile)
        write('GFXRECON_BEGIN_NAMESPACE(encode)', file=self.outFile)
        write('GFXRECON_BEGIN_NAMESPACE(vulkan_trackers)', file=self.outFile)
        self.newline()

    def endFile(self):
        """Method override."""

        self.newline()
        write('void* TrackStruct(const void* value, HandleUnwrapMemory* unwrap_memory)', file=self.outFile)
        write('{', file=self.outFile)
        write('    if (value == nullptr)', file=self.outFile)
        write('    {', file=self.outFile)
        write('        return nullptr;', file=self.outFile)
        write('    }', file=self.outFile)
        self.newline()
        write('    VkStructureType valueType = reinterpret_cast<const VkBaseInStructure*>(value)->sType;', file=self.outFile)
        write('    switch (valueType)', file=self.outFile)
        write('    {', file=self.outFile)
        for typename, struct_type_enum in self.struct_type_enums.items():
            write('        case {}:'.format(struct_type_enum), file=self.outFile)
            write('            return TrackStruct(reinterpret_cast<const {}*>(value), unwrap_memory);'.format(typename), file=self.outFile)
        write('        default:', file=self.outFile)
        write('            GFXRECON_LOG_ERROR("Unknown structure type: %u", valueType);', file=self.outFile)
        write('    }', file=self.outFile)
        self.newline()
        write('    return nullptr;', file=self.outFile)
        write('}', file=self.outFile)

        self.newline()
        write('GFXRECON_END_NAMESPACE(vulkan_trackers)', file=self.outFile)
        write('GFXRECON_END_NAMESPACE(encode)', file=self.outFile)
        write('GFXRECON_END_NAMESPACE(gfxrecon)', file=self.outFile)

        # Finish processing in superclass
        BaseGenerator.endFile(self)

    def genStruct(self, typeinfo, typename, alias):
        """Method override."""
        BaseGenerator.genStruct(self, typeinfo, typename, alias)

        if alias:
            return
        
        if typename in ['VkBaseInStructure', 'VkBaseOutStructure']:
            return
        
        struct_type_enum = self.make_structure_type_enum(typeinfo, typename)
        if struct_type_enum is None:
            return

        self.struct_type_enums[typename] = struct_type_enum
        
        write('{0}* TrackStruct(const {0}* value, HandleUnwrapMemory* unwrap_memory)'.format(typename), file=self.outFile)
        write('{', file=self.outFile)
        write('    if (value == nullptr)', file=self.outFile)
        write('    {', file=self.outFile)
        write('        return nullptr;', file=self.outFile)
        write('    }', file=self.outFile)
        self.newline()
        write('    {}* unwrapped_struct = vulkan_wrappers::MakeUnwrapStructs(value, 1, unwrap_memory);'.format(typename), file=self.outFile)
        self.newline()

        for value in self.feature_struct_members[typename]:
            if value.is_array and value.is_dynamic:
                if value.array_dimension == 1:
                    member_expr = f'unwrapped_struct->{value.name}'
                    length_expr = self.make_array_length_expression(value, 'unwrapped_struct->')
                    if value.base_type == 'void':
                        call_expr = f'vulkan_wrappers::MakeUnwrapStructs<uint8_t>(reinterpret_cast<const uint8_t*>({member_expr}), {length_expr}, unwrap_memory)'
                    else:
                        call_expr = f'vulkan_wrappers::MakeUnwrapStructs({member_expr}, {length_expr}, unwrap_memory)'
                    write('    if ({})'.format(member_expr), file=self.outFile)
                    write('    {', file=self.outFile)
                    write('        {} = {};'.format(member_expr, call_expr), file=self.outFile)
                    write('    }', file=self.outFile)
                else:
                    print(f'VulkanStructTrackersBodyGenerator ignored: {typename}.{value.name}') # TODO

        for value in self.feature_struct_members[typename]:
            if value.name == 'pNext':
                self.newline()
                write('    unwrapped_struct->pNext = TrackStruct(unwrapped_struct->pNext, unwrap_memory);', file=self.outFile)
                break
        
        self.newline()
        write('    return unwrapped_struct;', file=self.outFile)
        write('}', file=self.outFile)
        self.newline()




#!/usr/bin/python3 -i
#
# Copyright (c) 2021 Samsung
# Copyright (c) 2023 Google
# Copyright (c) 2023 LunarG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import sys
from base_generator import BaseGenerator, BaseGeneratorOptions, write
from vulkan_cpp_consumer_body_generator import \
    makeSnakeCaseName, makeGen, makeGenVar, makeGenVarCall, makeGenCond, makeGenConditions, \
    makeGenLoop, makeObjectType, makeOutStructSet, printOutStream

# Copyright text prefixing all headers (list of strings).
CPP_PREFIX_STRING = [
    '/*',
    '** Copyright (c) 2021 Samsung',
    '** Copyright (c) 2023 Google',
    '** Copyright (c) 2023 LunarG, Inc.',
    '**',
    '** Permission is hereby granted, free of charge, to any person obtaining a',
    '** copy of this software and associated documentation files (the "Software"),',
    '** to deal in the Software without restriction, including without limitation',
    '** the rights to use, copy, modify, merge, publish, distribute, sublicense,',
    '** and/or sell copies of the Software, and to permit persons to whom the',
    '** Software is furnished to do so, subject to the following conditions:',
    '**',
    '** The above copyright notice and this permission notice shall be included in',
    '** all copies or substantial portions of the Software.', '**',
    '** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR',
    '** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,',
    '** FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE',
    '** AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER',
    '** LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING',
    '** FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER',
    '** DEALINGS IN THE SOFTWARE.',
    '*/',
    '',
    '/*',
    '** This file is generated from the Khronos Vulkan XML API Registry.',
    '**',
    '*/',
    ''
]
BasicStringConversionHandledTypes = [
    'char',
    'int',
    'int32_t',
    'float',
    'double',
    'size_t',
    'uint32_t',
    'uint16_t',
    'uint64_t',
    'void',
    'VkBool32',
    'VkDeviceSize',
    'VkDeviceAddress'
]

class VulkanCppStructGeneratorOptions(BaseGeneratorOptions):
    """Options for generating a C++ class for Vulkan capture file to CPP structure generation"""

    def __init__(
        self,
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text=CPP_PREFIX_STRING,
        protect_file=False,
        protect_feature=True,
        extraVulkanHeaders=[]
    ):
        BaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            CPP_PREFIX_STRING,
            protect_file,
            protect_feature,
            extraVulkanHeaders=extraVulkanHeaders
        )

class VulkanCppStructGenerator(BaseGenerator):
    """VulkanCppStructGenerator - subclass of BaseGenerator.
    Generates vulkan struct generating functions.
    """

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
     ):
        BaseGenerator.__init__(
            self,
            process_cmds=True,
            process_structs=True,
            feature_break=False,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

        # Some structs are tracked like handles in the globalVar source so we only
        # need to properly initialize them, not define them.
        self.trackedStructs = [
            'VkMemoryRequirements2'
        ]

        self.overrideStructs = [
            'VkAndroidSurfaceCreateInfoKHR', # vkCreateAndroidSurfaceKHR, VK_KHR_ANDROID_SURFACE_EXTENSION_NAME
            'VkMetalSurfaceCreateInfoEXT',   # vkCreateMetalSurfaceEXT, VK_EXT_METAL_SURFACE_EXTENSION_NAME
            'VkWaylandSurfaceCreateInfoKHR', # vkCreateWaylandSurfaceKHR, VK_KHR_WAYLAND_SURFACE_EXTENSION_NAME
            'VkWin32SurfaceCreateInfoKHR',   # vkCreateWin32SurfaceKHR, VK_KHR_WIN32_SURFACE_EXTENSION_NAME
            'VkXcbSurfaceCreateInfoKHR',     # vkCreateXcbSurfaceKHR, VK_KHR_XCB_SURFACE_EXTENSION_NAME
            'VkXlibSurfaceCreateInfoKHR',    # vkCreateXlibSurfaceKHR, VK_KHR_XLIB_SURFACE_EXTENSION_NAME
        ]

        self.CUSTOM_GENERATE_STRUCT = [
            # Core API - Alphabetical
            'VkBaseInStructure',
            'VkBaseOutStructure',
            'VkBindSparseInfo',
            'VkClearColorValue',
            'VkDescriptorImageInfo',
            'VkDescriptorUpdateTemplateCreateInfoKHR',
            'VkDescriptorUpdateTemplateEntry',
            'VkInstanceCreateInfo',
            'VkMemoryAllocateFlagsInfo',
            'VkMemoryAllocateInfo',
            'VkPresentInfoKHR',
            'VkSubmitInfo',
            'VkSubmitInfo2',
            'VkSwapchainCreateInfoKHR',
            'VkWriteDescriptorSet',

            # EXT
            'VkDebugUtilsMessengerCreateInfoEXT',
            'VkImportMemoryHostPointerInfoEXT',

            # Other extensions
            'VkAccelerationStructureMotionInstanceNV',
            'VkImportAndroidHardwareBufferInfoANDROID',
        ]

        self.LOCAL_STRUCT_BLACKLIST = [
            'SECURITY_ATTRIBUTES',
            'GUID'
        ]

        self.next_index = 0

    def writeout(self, *args, **kwargs):
        write(*args, **kwargs, file=self.outFile)

    def newline(self):
        write('', file=self.outFile)

    def generateTodoFor(self, argName, indent):
        return printOutStream(['"// TODO: Support {argName} argument."'], locals(), indent=indent)

    # Method override
    def beginFile(self, genOpts):
        BaseGenerator.beginFile(self, genOpts)
        self.is_header = genOpts.filename.endswith(".h")

        if self.is_header:
            self.writeout('#include "util/defines.h"')
            self.includeVulkanHeaders(genOpts)
            self.newline()

        if not self.is_header:
            self.writeout('#include "decode/vulkan_cpp_structs.h"')
            self.writeout('#include "generated/generated_vulkan_cpp_structs.h"')
            self.writeout('#include "generated/generated_vulkan_cpp_consumer.h"')
            self.writeout('#include "generated/generated_vulkan_cpp_consumer_extension.h"')
            self.writeout('#include "generated/generated_vulkan_enum_to_string.h"')
        self.writeout('GFXRECON_BEGIN_NAMESPACE(gfxrecon)')
        self.writeout('GFXRECON_BEGIN_NAMESPACE(decode)')
        self.newline()

    # Method override
    def endFile(self):
        self.writeout('GFXRECON_END_NAMESPACE(decode)')
        self.writeout('GFXRECON_END_NAMESPACE(gfxrecon)')
        self.newline()

        # Finish processing in superclass
        BaseGenerator.endFile(self)

    def need_feature_generation(self):
        if self.struct_names:
            return True
        return False

    def is_union(self, value):
        if value not in self.registry.typedict:
            return False
        typeinfo = self.registry.typedict[value]
        type_elem = typeinfo.elem
        category = type_elem.get('category')
        if category == 'union':
            return True
        return False

    def generate_feature(self):
        structnames = self.get_filtered_struct_names()
        structnames.sort()

        # Insert the declaration of GenerateStruct functions.
        for structName in structnames:
            if (structName in self.CUSTOM_GENERATE_STRUCT or structName in self.feature_struct_aliases or structName in self.feature_union_aliases):
                continue

            struct_string = self.makeStructGenerateFunction(structName)
            struct_string += '\n'
            self.writeout(struct_string)

    #
    # Generate the "GenerateStruct_" function for the given structure name
    def makeStructGenerateFunction(self, struct_type):
        body = f'std::string GenerateStruct_{struct_type}(std::ostream &out, const {struct_type}* structInfo, ' \
                        f'Decoded_{struct_type}* metaInfo, VulkanCppConsumerBase &consumer)'
        if self.is_header:
            body += ';'
        else:
            body += '{\n'
            body += ''.join(self.makeStructDeclBody(struct_type))
            body += '}\n'
        return body

    def isStaticArray(self, array_length):
        if array_length[0].isupper() or array_length[0].isnumeric():
            return True
        else:
            return False

    def handlePNext(self, struct_prefix, indent, header, body, isFirstArg, isLastArg):
        local_header = []
        local_body = []
        local_header.extend(header)
        local_body.extend(body)

        local_header.append(makeGen(f'std::string pnext_name = GenerateExtension(out, {struct_prefix}pNext, metaInfo->pNext, consumer);', locals(), indent))
        local_body.append(makeOutStructSet('pnext_name', locals(), isFirstArg, isLastArg, indent))

        return local_header, local_body

    def handleChar(self, struct_prefix, arg, indent, header, body, isFirstArg, isLastArg):
        local_header = []
        local_body = []
        local_header.extend(header)
        local_body.extend(body)

        if arg.is_pointer and arg.pointer_count > 1:

            handleObjectType = None
            if arg.base_type in self.handle_names:
                handleObjectType = makeObjectType(arg.base_type)

            escapedStringArrayName = makeSnakeCaseName(f'{arg.name}Var'.format(**locals()))
            header_data = [
                makeGenVar(escapedStringArrayName, None, locals(), handleObjectType, indent, useThis=False),
                makeGenCond('{struct_prefix}{arg.array_length}',
                            [makeGenVar(escapedStringArrayName, '{arg.name}', handleObjectType, locals(), indent, addType=False, useThis=False),
                            printOutStream(['"const char* "',
                                            escapedStringArrayName,
                                            '"[] = "',
                                            f'VulkanCppConsumerBase::EscapeStringArray({struct_prefix}{arg.name}, {struct_prefix}{arg.array_length})',
                                            '";"'], locals(), indent)], [], locals(), indent)]
            local_header.append(''.join(header_data))
            local_body.append(makeOutStructSet(escapedStringArrayName, locals(), isFirstArg, isLastArg, indent))
        else:
            local_body.append(makeOutStructSet('VulkanCppConsumerBase::ToEscape({struct_prefix}{arg.name})', locals(), isFirstArg, isLastArg, indent))
        return local_header, local_body

    def handleBasicPointer(self, struct_prefix, arg, num_lengths, lengths, indent, header, body, isFirstArg, isLastArg):
        local_header = []
        local_body = []
        local_header.extend(header)
        local_body.extend(body)

        assert arg.pointer_count == 1, "Allow only a simple array"

        # ex: pDepthInputAttachmentIndex of VkRenderingInputAttachmentIndexInfoKHR has no count.
        if num_lengths == 0:
            return local_header, local_body

        # TODO: This is a hack, but according to the spec, this is correct
        if arg.name == 'pOffsetForRefFrame':
            lengths = []
            lengths.append(f'{struct_prefix}num_ref_frames_in_pic_order_cnt_cycle')
            num_lengths = 1

        handleObjectType = None
        if arg.base_type in self.handle_names:
            handleObjectType = makeObjectType(arg.base_type)

        variable_name = makeSnakeCaseName(arg.name + 'Array')

        this_header = [
            makeGenVar(variable_name, None, handleObjectType, locals(), indent, useThis=False),
            makeGenCond(f'{struct_prefix}{arg.name} != NULL', [
                makeGenVar(variable_name, arg.name, handleObjectType, locals(), indent + 4, addType=False, useThis=False),
                printOutStream([f'"{arg.base_type} "', '{variable_name} << "[] = "',
                                f'VulkanCppConsumerBase::BuildValue({struct_prefix}{arg.name}, {lengths[0]})',
                                '";"'], locals(), indent + 4)], [], locals(), indent)]
        local_header.extend(this_header)
        local_body.append(makeOutStructSet(variable_name, locals(), isFirstArg, isLastArg, indent))
        return local_header, local_body

    def handleLargeUint(self, struct_prefix, arg, num_lengths, lengths, indent, header, body, isFirstArg, isLastArg):
        local_header = []
        local_body = []
        local_header.extend(header)
        local_body.extend(body)

        if arg.is_array:
            strArrayName = makeSnakeCaseName(f'{arg.name}Array'.format(**locals()))
            strArrayValuesName = makeSnakeCaseName(f'{arg.name}Values'.format(**locals()))

            handleObjectType = None
            if arg.base_type in self.handle_names:
                handleObjectType = makeObjectType(arg.base_type)

            local_header.append(makeGenVar(strArrayName, strArrayName, handleObjectType, locals(), indent, addType=True, useThis=False)),
            local_header.append(makeGenCond('{lengths[0]} > 0', [
                makeGenVarCall('std::string', strArrayValuesName, 'toStringJoin',
                                [f'{struct_prefix}{arg.name}',
                                f'{struct_prefix}{arg.name} + {lengths[0]}',
                                f'[]({arg.base_type} current) {{{{ return std::to_string(current); }}}}',
                                '", "'], locals(), indent + 4),
                makeGenConditions([
                    ['{lengths[0]} == 1', [makeGen(f'{strArrayName} = "&" + {strArrayValuesName};', locals(), indent + 8)]],
                    ['{lengths[0]} > 1', [printOutStream([f'"{arg.base_type} "', strArrayName, '"[] = {{"', strArrayValuesName, '"}};"'], locals(), indent + 8)]],
                    # No need for else case
                ], locals(), indent + 4),
            ], [], locals(), indent))
            local_body.append(makeOutStructSet(f'"{{{{ *" << {strArrayName} << " }}}}"', locals(), isFirstArg, isLastArg, indent))
        else:
            local_body.append(makeOutStructSet(f'{struct_prefix}{arg.name} << "UL"', locals(), isFirstArg, isLastArg, indent))

        return local_header, local_body

    def handleEnum(self, struct_prefix, arg, num_lengths, lengths, indent, header, body, isFirstArg, isLastArg):
        local_header = []
        local_body = []
        local_header.extend(header)
        local_body.extend(body)

        if arg.is_pointer:
            # TODO: Possibly an array of enum/flag values?
            if num_lengths > 0:
                strArrayName = makeSnakeCaseName(f'{arg.name}Array'.format(**locals()))
                strArrayValuesName = makeSnakeCaseName(f'{arg.name}Values'.format(**locals()))

                handleObjectType = None
                if arg.base_type in self.handle_names:
                    handleObjectType = makeObjectType(arg.base_type)

                argHandler = [
                    makeGen(f'std::string {strArrayValuesName};', locals(), indent),
                    makeGenVar(strArrayName, None, handleObjectType, locals(), indent, useThis=False),
                    makeGenCond(f'{struct_prefix}{arg.name} != NULL', [
                        makeGenLoop('idx', '{lengths[0]}', [
                            makeGen(f'{strArrayValuesName} += util::ToString<{arg.base_type}>({struct_prefix}{arg.name}[idx]) + ", ";', locals(), indent + 8),
                        ], locals(), indent + 4),
                        makeGenVar(strArrayName, arg.name, handleObjectType, locals(), indent + 4, addType=False, useThis=False),
                        printOutStream([f'"{arg.base_type} "', strArrayName, '"[] = {{"', strArrayValuesName, '"}};"'], locals(), indent + 4)
                    ], [], locals(), indent),
                ]
                local_header.append(''.join(argHandler))
                local_body.append(makeOutStructSet(strArrayName, locals(), isFirstArg, isLastArg, indent))
            else:
                local_body.append(self.generateTodoFor(arg.name + '(Enum/flag pointer)', indent))
        else:
            local_body.append(makeOutStructSet(f'"{arg.base_type}(" << {struct_prefix}{arg.name} << ")"', locals(), isFirstArg, isLastArg, indent))

        return local_header, local_body


    def handleHandle(self, struct_prefix, arg, num_lengths, lengths, indent, header, body, isFirstArg, isLastArg):
        local_header = []
        local_body = []
        local_header.extend(header)
        local_body.extend(body)

        if arg.is_array:
            # Generate an array of strings containing the handle id names
            strArrayName = makeSnakeCaseName(arg.name + "Array")
            strArrayValuesName = makeSnakeCaseName(arg.name + "Values")

            handleObjectType = None
            if arg.base_type in self.handle_names:
                handleObjectType = makeObjectType(arg.base_type)

            local_header.append(makeGenVar(strArrayName, None, handleObjectType, locals(), indent, useThis=False))
            local_header.append(makeGenCond(f'metaInfo->{arg.name}.GetPointer() != NULL && {lengths[0]} > 0', [
                makeGenVar(strArrayName, strArrayName, handleObjectType, locals(), indent + 4, addType=False, useThis=False),
                makeGenVarCall('std::string', strArrayValuesName, 'toStringJoin',
                                [f'metaInfo->{arg.name}.GetPointer()',
                                f'metaInfo->{arg.name}.GetPointer() + {lengths[0]}',
                                '[&](const format::HandleId current) {{ return consumer.GetHandle(current); }}',
                                '", "'], locals(), indent + 4),
                makeGenConditions([
                    ['{lengths[0]} == 1', [makeGen(f'{strArrayName} = "&" + {strArrayValuesName};', locals(), indent + 8)]],
                    ['{lengths[0]} > 1', [printOutStream([f'"{arg.base_type} "', strArrayName, '"[] = {{"', strArrayValuesName, '"}};"'], locals(), indent + 8)]],
                    # No need for else case
                ], locals(), indent + 4),
            ], [], locals(), indent))

            local_body.append(makeOutStructSet(strArrayName, locals(), isFirstArg, isLastArg, indent))
        else:
            local_body.append(makeOutStructSet(f'consumer.GetHandle(metaInfo->{arg.name})', locals(), isFirstArg, isLastArg, indent))

        return local_header, local_body

    def handleInputArray(self, struct_prefix, arg, num_lengths, lengths, indent, header, body, isFirstArg, isLastArg):
        local_header = []
        local_body = []
        local_header.extend(header)
        local_body.extend(body)

        structBuild = []
        type_cast_prefix = ''
        type_cast_suffix = ''
        if arg.base_type == 'void':
            arg.base_type = 'uint8_t'
            type_cast_prefix = 'reinterpret_cast<const uint8_t*>('
            type_cast_suffix = ')'

        handleObjectType = None
        if arg.base_type in self.handle_names:
            handleObjectType = makeObjectType(arg.base_type)

        if num_lengths > 0:
            strArrayName = makeSnakeCaseName(arg.name + "Array")
            strArrayValuesName = makeSnakeCaseName(arg.name + "Values")

            structBuild += makeGenVar(strArrayName, None, handleObjectType, locals(), indent, useThis=False)
            space = (' ' * indent)
            
            # If pointer and not just static array
            if not self.isStaticArray(lengths[0]):
                structBuild += f'{space}if ({struct_prefix}{arg.name} != NULL) {{\n'
                indent += 4
                space = (' ' * indent)

            structBuild += makeGen('std::string {strArrayValuesName};', locals(), indent)

            var_suffix = ''
            for count, cur_length in enumerate(lengths):
                space = (' ' * indent)
                structBuild += f'{space}for (uint32_t idx{count} = 0; idx{count} < {cur_length}; ++idx{count}) {{\n'
                var_suffix = var_suffix + f'[idx{count}]'
                indent = indent + 4
                if count < num_lengths - 1:
                    space = (' ' * indent)
                    structBuild += f'{space}{strArrayValuesName} += "{{ ";\n'

            cur_arg_value = f'{type_cast_prefix}{struct_prefix}{arg.name}{type_cast_suffix}{var_suffix}'
            structBuild += makeGen(f'{strArrayValuesName} += std::to_string({cur_arg_value}) + ", ";', locals(), indent)

            for count, cur_length in enumerate(lengths):
                if count < num_lengths - 1:
                    space = (' ' * indent)
                    structBuild += f'{space}{strArrayValuesName} += " }},";\n'

                indent = indent - 4
                space = (' ' * indent)
                structBuild += f'{space}}}\n'

            structBuild += makeGenVar(strArrayName, arg.name, handleObjectType, locals(), indent, addType=False, useThis=False)
            structBuild += printOutStream([f'"{arg.base_type} "', strArrayName, '"[] = {{"', strArrayValuesName, '"}};"'], locals(), indent)

            if not self.isStaticArray(lengths[0]):
                indent = indent - 4
                space = (' ' * indent)
                structBuild += f'{space}}}\n'

            local_header.append(''.join(structBuild))
            local_body.append(makeOutStructSet(strArrayName, locals(), isFirstArg, isLastArg, indent))

        else:
            local_body.append(self.generateTodoFor(arg.name + '(input pointer)', indent))

        return local_header, local_body
    

    def handleOutputParam(self, struct_prefix, arg, num_lengths, lengths, indent, header, body, isFirstArg, isLastArg):
        local_header = []
        local_body = []
        local_header.extend(header)
        local_body.extend(body)

        handleObjectType = None
        if arg.base_type in self.handle_names:
            handleObjectType = makeObjectType(arg.base_type)

        # TODO: is the output parameter okay here?
        if not arg.is_array:
            local_body.append(self.generateTodoFor(arg.name + " (output?)", indent))
        else:
            if arg.array_length_value is None:
                if lengths[0].isnumeric() or lengths[0].isupper():
                    if len(lengths):
                        param_prefix = '&'
                        param_suffix = '[0]' * len(lengths)
                        struct_param = f'reinterpret_cast<const {arg.base_type}*>({param_prefix}{struct_prefix}{arg.name}{param_suffix})'
                    else:
                        struct_param = f'reinterpret_cast<const {arg.base_type}*>({struct_prefix}{arg.name})'
                else:
                    struct_param = f'{struct_prefix}{arg.name}'

                local_body.append(makeOutStructSet(f'VulkanCppConsumerBase::BuildValue({struct_param}, {lengths[0]})', locals(), isFirstArg, isLastArg, indent))
            else: # arg.array_length_value
                if self.is_struct(arg.base_type):
                    innerIndent = indent + 4 if not arg.array_capacity else 4

                    varName = makeSnakeCaseName(arg.name + "Names")
                    arrayVarName = makeSnakeCaseName(arg.name + "Array")
                    valuesVarName = makeSnakeCaseName(arg.name + "Values")

                    arrayBuilder = [
                        makeGen(f'std::string {varName};', locals(), indent=innerIndent),
                        makeGenLoop('idx', f'{lengths[0]}', [
                            makeGen('std::string variable_name = "NULL";', indent=innerIndent + 4),
                            makeGenCond(f'{struct_prefix}{arg.name} + idx != NULL', [
                                makeGenVarCall(None,
                                                'variable_name',
                                                f'GenerateStruct_{arg.base_type}',
                                                ['out',
                                                f'{struct_prefix}{arg.name} + idx',
                                                f'metaInfo->{arg.name}->GetMetaStructPointer() + idx',
                                                'consumer'],
                                                locals(), indent=innerIndent + 8)
                            ], [], locals(), indent=innerIndent + 4),
                            makeGen('{varName} += variable_name + ", ";', locals(), indent=innerIndent + 4),
                        ], locals(), indent=innerIndent),
                        printOutStream([f'"{arg.base_type} "', arrayVarName, '"[] = {{"', varName, '"}};"'],
                                        locals(), indent=innerIndent)
                    ]

                    if arg.array_capacity:
                        local_body.append('\tstruct_body << "\\t\\t\\t{}," << std::endl;\n')
                    else:
                        arrayName = makeGenVar(arrayVarName, None, handleObjectType, locals(), indent, useThis=False)
                        arrayNameSet = makeGenVar(arrayVarName, arg.name, handleObjectType, locals(), indent + 4, addType=False, useThis=False)
                        arrayProcess = makeGenCond(f'{struct_prefix}{arg.name} != NULL', [arrayNameSet] + arrayBuilder, [], locals(), indent)

                        local_header.append(arrayName + arrayProcess)
                        local_body.append(makeOutStructSet(arrayVarName, locals(), isFirstArg, isLastArg, indent))
                elif self.is_enum(arg.base_type) or self.is_flags(arg.base_type):
                    arrayVarName = makeSnakeCaseName(arg.name + "Array")
                    valuesVarName = makeSnakeCaseName(arg.name + "Values")

                    argHandler = [
                        makeGenVar(arrayVarName, None, handleObjectType, locals(), indent, useThis=False),
                        makeGenCond(f'{struct_prefix}{arg.name} != NULL', [
                            makeGen(f'std::string {valuesVarName};', locals(), indent + 4),
                            makeGenLoop('idx', '{lengths[0]}', [
                                makeGen(f'{valuesVarName} += util::ToString<{arg.base_type}>({struct_prefix}{arg.name}[idx]) + ", ";', locals(), indent + 8),
                            ], locals(), indent + 4),
                            makeGenVar(arrayVarName, '{arg.name}', handleObjectType, locals(), indent + 4, addType=False, useThis=False),
                            printOutStream([f'"{arg.base_type} "', arrayVarName, '"[] = {{"', valuesVarName, '"}};"'], locals(), indent + 4)
                        ], [], locals(), indent),
                    ]
                    local_header.append(''.join(argHandler))
                    local_body.append(makeOutStructSet(arrayVarName, locals(), isFirstArg, isLastArg, indent))
                else:
                    local_body.append(self.generateTodoFor(arg.name + " (output with array length value?)", indent))

        return local_header, local_body

    #
    # Generate structure declaration body.
    def makeStructDeclBody(self, structName):
        structMembers = self.feature_struct_members[structName]
        vkLUType = ['VkDeviceSize', 'VkDeviceAddress', 'uint64_t']
        indent = 4
        body = []
        header = []
        header.append(makeGen('std::stringstream struct_body;', locals(), indent))
        struct_prefix = 'structInfo->'

        for arg in structMembers:
            isFirstArg = (arg == structMembers[0])
            isLastArg = (arg == structMembers[-1])

            if arg.base_type == 'VkDescriptorUpdateTemplateEntryKHR':
                print(f'   {arg.name} was struct base type {arg.base_type} should have an alias')

            if arg.base_type in self.feature_struct_aliases:
                print(f'   {arg.name} was struct base type {arg.base_type} but aliased to {self.feature_struct_aliases[arg.base_type]}')
                arg.base_type = self.feature_struct_aliases[arg.base_type]
            elif arg.base_type in self.feature_union_aliases:
                print(f'   {arg.name} was union base type {arg.base_type} but aliased to {self.feature_union_aliases[arg.base_type]}')
                arg.base_type = self.feature_union_aliases[arg.base_type]

            handleObjectType = None
            if arg.base_type in self.handle_names:
                handleObjectType = makeObjectType(arg.base_type)

            arg_name = struct_prefix + arg.name

            lengths = []
            if arg.array_length_value is not None:
                loopCondition = '{arg.array_length}'.format(**locals())
                varNameArrayLength = getattr(arg.array_length_value, 'name', None)
                # Substitute the variable occurrence with the proper variable reference.
                # This is required if the size of the array is an expression.
                if varNameArrayLength:
                    loopCondition = re.sub(varNameArrayLength, '{struct_prefix}{varNameArrayLength}'.format(**locals()), loopCondition)
                    lengths = re.split(',', loopCondition)
            elif arg.array_length is not None:
                lengths = re.split(',', arg.array_length)
                for index, length in enumerate(lengths):
                    if not (length.isupper() or length.isnumeric()):
                        lengths[index] = struct_prefix + length

            num_lengths = len(lengths)

            if arg.name == 'pNext':
                header, body = self.handlePNext(struct_prefix, indent, header, body, isFirstArg, isLastArg)

            elif arg.base_type == 'char':
                header, body = self.handleChar(struct_prefix, arg, indent, header, body, isFirstArg, isLastArg)

            elif arg.base_type in ['float', 'int32_t', 'uint32_t', 'VkClearValue', 'VkRect2D'] and arg.is_pointer:
                header, body = self.handleBasicPointer(struct_prefix, arg, num_lengths, lengths, indent, header, body, isFirstArg, isLastArg)

            elif arg.base_type in vkLUType:
                header, body = self.handleLargeUint(struct_prefix, arg, num_lengths, lengths, indent, header, body, isFirstArg, isLastArg)

            elif self.is_output_parameter(arg):
                header, body = self.handleOutputParam(struct_prefix, arg, num_lengths, lengths, indent, header, body, isFirstArg, isLastArg)

            elif self.is_union(arg.base_type):
                if not arg.is_pointer:
                    if arg.base_type in ['VkClearValue', 'VkClearColorValue', 'VkComponentMapping', 'VkOffset2D', 'VkExtent2D']:
                        body.append(makeOutStructSet('VulkanCppConsumerBase::BuildValue({arg_name})', locals(), isFirstArg, isLastArg, indent))
                    elif arg.base_type in self.feature_union_members:
                        union_members = self.feature_union_members[arg.base_type]
                        first_member = union_members[0]

                        if first_member.base_type in BasicStringConversionHandledTypes:
                            body.append(makeOutStructSet(f'{arg_name}.{first_member.name}', locals(), isFirstArg, isLastArg, indent))
                        elif self.is_struct(first_member.base_type):
                            varName = makeSnakeCaseName(first_member.name + "InfoVar")

                            argHandler = [
                                makeGenVarCall('std::string',
                                                varName,
                                                f'GenerateStruct_{first_member.base_type}',
                                                ['out',
                                                f'&{arg_name}.{first_member.name}',
                                                f'metaInfo->{arg.name}->{first_member.name}',
                                                'consumer'],
                                                locals(), indent),
                            ]
                            header.append(''.join(argHandler))

                            body.append(makeOutStructSet(varName, locals(), isFirstArg, isLastArg, indent))
                        else:
                            body.append(makeOutStructSet('VulkanCppConsumerBase::BuildValue({arg_name}.{first_member.name})', locals(), isFirstArg, isLastArg, indent))
                    else:
                        body.append(makeOutStructSet('VulkanCppConsumerBase::BuildValue({arg_name})', locals(), isFirstArg, isLastArg, indent))
                else:
                    #if arg.base_type in ['VkClearColorValue', 'VkComponentMapping', 'VkOffset2D', 'VkExtent2D']:
                    body.append(makeGenVar(f'tmp{arg.name}', arg.name, handleObjectType, locals(), indent, useThis=False))
                    body.append(makeOutStructSet('VulkanCppConsumerBase::BuildValue({arg_name})', locals(), isFirstArg, isLastArg, indent + 4))
            elif self.is_struct(arg.base_type) and arg.base_type not in self.LOCAL_STRUCT_BLACKLIST:
                if not arg.is_pointer:
                    varName = makeSnakeCaseName(arg.name + "InfoVar")
                    argHandler = [
                        makeGenVarCall('std::string',
                                        varName,
                                        f'GenerateStruct_{arg.base_type}',
                                        ['out',
                                        f'&{arg_name}',
                                        f'metaInfo->{arg.name}',
                                        'consumer'],
                                        locals(), indent),
                    ]
                    header.append(''.join(argHandler))

                    body.append(makeOutStructSet(varName, locals(), isFirstArg, isLastArg, indent))
                else:
                    if num_lengths > 0:
                        arrayVarName = makeSnakeCaseName(arg.name + "Array")
                        namesVarName = makeSnakeCaseName(arg.name + "Names")

                        if num_lengths > 1:

                            indent = 4
                            structBuild = makeGenVar(arrayVarName, None, handleObjectType, locals(), indent, useThis=False)
                            space = (' ' * indent)
                            structBuild += f'{space}if ({arg_name} != NULL) {{\n'
                            indent = 8
                            structBuild += makeGenVar(arrayVarName, arg.name, handleObjectType, locals(), indent, addType=False, useThis=False)
                            structBuild += makeGen(f'std::string {namesVarName};', locals(), indent)

                            member_prefix = '&('
                            member_suffix = ')'
                            still_is_pointer = False
                            if arg.pointer_count > len(lengths):
                                still_is_pointer = True
                                member_prefix = ''
                                member_suffix = ''

                            var_suffix = ''
                            for count, cur_length in enumerate(lengths):
                                space = (' ' * indent)
                                structBuild += f'{space}for (uint32_t idx{count} = 0; idx{count} < {cur_length}; ++idx{count}) {{\n'
                                var_suffix = var_suffix + f'[idx{count}]'
                                indent = indent + 4

                            structBuild += makeGen('std::string variable_name = "NULL";', locals(), indent)
                            space = (' ' * indent)
                            
                            if still_is_pointer:
                                structBuild += f'{space}{arg_name}{var_suffix} != NULL) {{'
                                indent = indent + 4

                            structBuild += makeGenVarCall(None,
                                                    'variable_name',
                                                    f'GenerateStruct_{arg.base_type}',
                                                    ['out',
                                                        f'{member_prefix}{arg_name}{var_suffix}{member_suffix}',
                                                        f'{member_prefix}metaInfo->{arg.name}->GetMetaStructPointer(){var_suffix}{member_suffix}',
                                                        'consumer'], locals(), indent)
                            structBuild += makeGen(f'{namesVarName} += variable_name + ", ";', locals(), indent)

                            if still_is_pointer:
                                indent = indent - 4
                                space = (' ' * indent)
                                structBuild += f'{space}}}\n'

                            for count, cur_length in enumerate(lengths):
                                indent = indent - 4
                                space = (' ' * indent)
                                structBuild += f'{space}}}\n'

                            indent = indent - 4
                            space = (' ' * indent)
                            structBuild += f'{space}}}\n'
                        else:
                            structBuild = [
                                makeGenVar(arrayVarName, None, handleObjectType, locals(), indent, useThis=False),
                                makeGenCond(f'{arg_name} != NULL', [
                                    makeGenVar(arrayVarName, arg.name, handleObjectType, locals(), indent + 4, addType=False, useThis=False),
                                    makeGen(f'std::string {namesVarName};', locals(), indent + 4),
                                    makeGenLoop('idx', f'{lengths[0]}', [
                                        makeGen('std::string variable_name = "NULL";', locals(), indent + 8),
                                        makeGenCond(f'{arg_name} + idx != NULL', [
                                            makeGenVarCall(None,
                                                        'variable_name',
                                                        f'GenerateStruct_{arg.base_type}',
                                                        ['out',
                                                            f'{arg_name} + idx',
                                                            f'metaInfo->{arg.name}->GetMetaStructPointer() + idx',
                                                            'consumer'],
                                                        locals(), indent + 12),
                                        ], [], locals(), indent + 8),
                                        makeGen(f'{namesVarName} += variable_name + ", ";', locals(), indent + 8),
                                    ], locals(), indent + 4),
                                    printOutStream([f'"{arg.base_type} "', arrayVarName, '"[] = {{"', namesVarName, '"}};"'], locals(), indent + 4),
                                ], [], locals(), indent)
                            ]
                        header.append(''.join(structBuild))
                        body.append(makeOutStructSet(arrayVarName, locals(), isFirstArg, isLastArg, indent))
                    else:
                        varName = makeSnakeCaseName(arg.name + "Struct")

                        prefixVar = f'prefix_{arg.name}'
                        # Make sure if the sub struct pointer during capture processing does contain data and it is not a NULL value.
                        structBuild = [
                            makeGenVar(varName, None, handleObjectType, locals(), indent, useThis=False),
                            makeGenCond('{arg_name} != NULL', [
                                makeGenVarCall(None,
                                               varName,
                                               f'GenerateStruct_{arg.base_type}',
                                               ['out',
                                                f'{arg_name}',
                                                f'metaInfo->{arg.name}->GetMetaStructPointer()',
                                                'consumer'],
                                               locals(), indent + 4),
                                makeGen('{varName}.insert(0, "&");', locals(), indent + 4),
                            ], [], locals(), indent),
                        ]
                        header.append(''.join(structBuild))
                        body.append(makeOutStructSet(varName, locals(), isFirstArg, isLastArg, indent))
                # TODO:
                # structArguments = []
                # for subArg in self.feature_struct_members[arg.base_type]:
                #     structArguments.append('{struct_prefix}{structName}.{structMemberName}\n'.format(structName=arg.name, structMemberName=subArg.name))
                # body += '    out << varname << ".{argName} = {{ " << {structArgs}    << "}};" << std::endl;\n'.format(argName=arg.name, structArgs='    << "," << '.join(structArguments))

            elif self.is_enum(arg.base_type) or self.is_flags(arg.base_type):
                header, body = self.handleEnum(struct_prefix, arg, num_lengths, lengths, indent, header, body, isFirstArg, isLastArg)

            elif self.is_handle(arg.base_type):
                header, body = self.handleHandle(struct_prefix, arg, num_lengths, lengths, indent, header, body, isFirstArg, isLastArg)

            elif self.is_input_pointer(arg) and arg.is_array:
                header, body = self.handleInputArray(struct_prefix, arg, num_lengths, lengths, indent, header, body, isFirstArg, isLastArg)

            else:
                struct_arg = arg_name
                if 'int8_t' in arg.base_type:
                    struct_arg = f'std::to_string({arg_name})'
                body.append(makeOutStructSet(struct_arg, locals(), isFirstArg, isLastArg, indent))

        if structName in self.feature_struct_aliases:
            structName = self.feature_struct_aliases[structName]

        var_list = []

        # If we already have a variable we're going to overwrite, just set the variable = to the new values.
        # Otherwise, we need to define a new variable
        if structName in self.trackedStructs:
            var_list.append('variable_name')
            var_list.append('" = {{"')
        else:
            var_list.append(f'"{structName} "')
            var_list.append('variable_name')
            var_list.append('" {{"')

        # convert name: Vk<StructName> => <structName>
        # TODO: maybe improve this a bit?
        structVarName = structName
        if structVarName.startswith("Vk"):
            structVarName = structVarName[2:]

        structVarName = structVarName[0].lower() + structVarName[1:]

        # insert the header up front (where all the variables are defined)
        body.append(makeGen(f'std::string variable_name = consumer.AddStruct(struct_body, "{structVarName}");', locals(), indent))
        body.append(printOutStream(var_list, locals(), indent))
        body.append(printOutStream(['struct_body.str()'], locals(), indent))
        body.append(printOutStream(['"}};"'], locals(), indent))

        if structName in self.overrideStructs:
            body.append(printOutStream(['"Override{structName}(&"', 'variable_name', '", "', '"appdata"', '");"'], locals(), indent))

        body.append(makeGen('return variable_name;', locals(), indent))

        func = []
        func.extend(header)
        func.extend(body)
        return func



#!/usr/bin/python3 -i
#
# Copyright (c) 2021-2022 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import re
import sys
import inspect
from base_generator import *


class VulkanEnumToStringBodyGeneratorOptions(BaseGeneratorOptions):
    """Options for generating C++ functions for Vulkan ToString() functions"""

    def __init__(
        self,
        # Path to JSON file listing apicalls and structs to ignore.
        blacklists=None,
        # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        platform_types=None,
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extraVulkanHeaders=[]
    ):
        BaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extraVulkanHeaders=extraVulkanHeaders
        )


# VulkanEnumToStringBodyGenerator - subclass of BaseGenerator.
# Generates C++ functions for stringifying Vulkan API enums.
class VulkanEnumToStringBodyGenerator(BaseGenerator):
    """Generate C++ functions for Vulkan ToString() functions"""

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        BaseGenerator.__init__(
            self,
            process_cmds=False,
            process_structs=True,
            feature_break=True,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

        # Set of enums that have been processed since we'll encounter enums that are
        #   referenced by extensions multiple times.  This list is prepopulated with
        #   enums that should be skipped.
        self.processedEnums = {
            'VkAccessFlagBits2KHR',
            'VkPipelineStageFlagBits2KHR',
        }

    # Method override
    # yapf: disable
    def beginFile(self, genOpts):
        BaseGenerator.beginFile(self, genOpts)
        body = inspect.cleandoc('''
            #include "generated_vulkan_enum_to_string.h"

            GFXRECON_BEGIN_NAMESPACE(gfxrecon)
            GFXRECON_BEGIN_NAMESPACE(util)
            ''')
        write(body, file=self.outFile)
    # yapf: enable

    # Method override
    # yapf: disable
    def endFile(self):
        body = inspect.cleandoc('''
            GFXRECON_END_NAMESPACE(util)
            GFXRECON_END_NAMESPACE(gfxrecon)
            ''')
        write(body, file=self.outFile)

        # Finish processing in superclass
        BaseGenerator.endFile(self)
    # yapf: enable

    #
    # Indicates that the current feature has C++ code to generate.
    def need_feature_generation(self):
        self.feature_break = False
        if self.feature_struct_members:
            return True
        return False

    # Performs C++ code generation for the feature.
    # yapf: disable

    def generate_feature(self):
        for enum in sorted(self.enum_names):
            if not enum in self.processedEnums and not enum in self.enumAliases:
                self.processedEnums.add(enum)
                if self.is_flags_enum_64bit(enum):
                    # print(enum)
                    # body = 'std::string {0}ToString(const {0}& value, ToStringFlags, uint32_t, uint32_t)\n'
                    # Since every caller needs to know exactly what it is calling, we may as well
                    # dispense with the parameters that are always ignored:
                    body = 'std::string {0}ToString(const {0} value)\n'
                else:
                    body = 'template <> std::string ToString<{0}>(const {0}& value, ToStringFlags, uint32_t, uint32_t)\n'
                body += '{{\n'
                enumerants = self.enumEnumerants[enum]
                if len(enumerants):
                    body += '    switch (value) {{\n'
                    for enumerant in enumerants:
                        body += '    case {0}: return "{0}";\n'.format(
                            enumerant)
                    body += '    default: break;\n'
                    body += '    }}\n'
                body += '    return "Unhandled {0}";\n'
                body += '}}\n'
                if 'Bits' in enum:
                    if self.is_flags_enum_64bit(enum):
                        body += '\nstd::string {1}ToString(VkFlags64 vkFlags)\n'
                        body += '{{\n'
                        body += '    std::string   str;\n'
                        body += '    VkFlags64     index = 0U;\n'
                        body += '    while (vkFlags)\n'
                        body += '    {{\n'
                        body += '        if (vkFlags & 1U)\n'
                        body += '        {{\n'
                        body += '            if (!str.empty())\n'
                        body += '            {{\n'
                        body += '                str += \'|\';\n'
                        body += '            }}\n'
                        body += '            str.append({0}ToString(static_cast<{0}>(1U) << index));\n'
                        body += '        }}\n'
                        body += '        ++index;\n'
                        body += '        vkFlags >>= 1U;\n'
                        body += '    }}\n'
                        body += '    if (str.empty())\n'
                        body += '    {{\n'
                        body += '        str.append({0}ToString(0U));\n'
                        body += '    }}\n'
                        body += '    return str;\n'
                        body += '}}\n'
                    else:
                        # Original version(these are never actually being called which is part of issue #620):
                        body += '\ntemplate <> std::string ToString<{0}>(VkFlags vkFlags, ToStringFlags, uint32_t, uint32_t)\n'
                        # Simpler, non-template version that matches the 64 bit version above. Changing
                        # to these signatures actually compiles fine, showing the originals were never
                        # called anywhere. Leaving this commented-out but ready for the PR that fixes
                        # issue #620 to use.
                        # body += '\nstd::string {1}ToString(VkFlags vkFlags)\n'
                        body += '{{\n'
                        body += '    return BitmaskToString<{0}>(vkFlags);\n'
                        body += '}}\n'
                write(body.format(enum, BitsEnumToFlagsTypedef(enum)),
                      file=self.outFile)
                # if self.is_flags_enum_64bit(enum):
                #    print(body.format(enum, BitsEnumToFlagsTypedef(enum)))

    # yapf: enable



#!/usr/bin/env python
#
# Copyright (c) 2018-2023 Valve Corporation
# Copyright (c) 2018-2023 LunarG, Inc.
# Copyright (c) 2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# The content of this file was derived from the Khronos Registry gencode.py
# and related Python files found in the KhronosGroup/Vulkan-Headers GitHub repository.
#
# Copyright (c) 2013-2016 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import pdb
import sys
import time
from registry.reg import Registry, etree
from generator import write

# API Call Decoders
from vulkan_decoder_body_generator import VulkanDecoderBodyGenerator, VulkanDecoderBodyGeneratorOptions
from vulkan_decoder_header_generator import VulkanDecoderHeaderGenerator, VulkanDecoderHeaderGeneratorOptions

# Struct Decoders
from vulkan_struct_decoders_body_generator import VulkanStructDecodersBodyGenerator, VulkanStructDecodersBodyGeneratorOptions
from vulkan_struct_decoders_forward_generator import VulkanStructDecodersForwardGenerator, VulkanStructDecodersForwardGeneratorOptions
from vulkan_struct_decoders_header_generator import VulkanStructDecodersHeaderGenerator, VulkanStructDecodersHeaderGeneratorOptions
from decode_pnext_struct_generator import DecodePNextStructGenerator, DecodePNextStructGeneratorOptions

# Consumers
from vulkan_consumer_header_generator import VulkanConsumerHeaderGenerator, VulkanConsumerHeaderGeneratorOptions
from vulkan_cpp_consumer_body_generator import VulkanCppConsumerBodyGenerator,VulkanCppConsumerBodyGeneratorOptions
from vulkan_cpp_consumer_header_generator import VulkanCppConsumerHeaderGenerator, VulkanCppConsumerHeaderGeneratorOptions
from vulkan_json_consumer_header_generator import VulkanExportJsonConsumerHeaderGenerator, VulkanExportJsonConsumerHeaderGeneratorOptions
from vulkan_json_consumer_body_generator import VulkanExportJsonConsumerBodyGenerator, VulkanExportJsonConsumerBodyGeneratorOptions
from vulkan_replay_consumer_body_generator import VulkanReplayConsumerBodyGenerator, VulkanReplayConsumerBodyGeneratorOptions
from vulkan_referenced_resource_consumer_header_generator import VulkanReferencedResourceHeaderGenerator, VulkanReferencedResourceHeaderGeneratorOptions
from vulkan_referenced_resource_consumer_body_generator import VulkanReferencedResourceBodyGenerator, VulkanReferencedResourceBodyGeneratorOptions
from vulkan_struct_handle_mappers_header_generator import VulkanStructHandleMappersHeaderGenerator, VulkanStructHandleMappersHeaderGeneratorOptions
from vulkan_struct_handle_mappers_body_generator import VulkanStructHandleMappersBodyGenerator, VulkanStructHandleMappersBodyGeneratorOptions
from vulkan_feature_util_body_generator import VulkanFeatureUtilBodyGenerator, VulkanFeatureUtilBodyGeneratorOptions

# API Call Encoders
from vulkan_api_call_encoders_body_generator import VulkanApiCallEncodersBodyGenerator, VulkanApiCallEncodersBodyGeneratorOptions
from vulkan_api_call_encoders_header_generator import VulkanApiCallEncodersHeaderGenerator, VulkanApiCallEncodersHeaderGeneratorOptions
from vulkan_command_buffer_util_body_generator import VulkanCommandBufferUtilBodyGenerator, VulkanCommandBufferUtilBodyGeneratorOptions
from vulkan_command_buffer_util_header_generator import VulkanCommandBufferUtilHeaderGenerator, VulkanCommandBufferUtilHeaderGeneratorOptions
from vulkan_dispatch_table_generator import VulkanDispatchTableGenerator, VulkanDispatchTableGeneratorOptions
from layer_func_table_generator import LayerFuncTableGenerator, LayerFuncTableGeneratorOptions

# Struct Encoders
from vulkan_struct_encoders_body_generator import VulkanStructEncodersBodyGenerator, VulkanStructEncodersBodyGeneratorOptions
from vulkan_struct_encoders_header_generator import VulkanStructEncodersHeaderGenerator, VulkanStructEncodersHeaderGeneratorOptions
from encode_pnext_struct_generator import EncodePNextStructGenerator, EncodePNextStructGeneratorOptions
from vulkan_struct_handle_wrappers_header_generator import VulkanStructHandleWrappersHeaderGenerator, VulkanStructHandleWrappersHeaderGeneratorOptions
from vulkan_struct_handle_wrappers_body_generator import VulkanStructHandleWrappersBodyGenerator, VulkanStructHandleWrappersBodyGeneratorOptions
from vulkan_struct_trackers_header_generator import VulkanStructTrackersHeaderGenerator, VulkanStructTrackersHeaderGeneratorOptions
from vulkan_struct_trackers_body_generator import VulkanStructTrackersBodyGenerator, VulkanStructTrackersBodyGeneratorOptions

# To String
from vulkan_enum_to_string_body_generator import VulkanEnumToStringBodyGenerator, VulkanEnumToStringBodyGeneratorOptions
from vulkan_enum_to_string_header_generator import VulkanEnumToStringHeaderGenerator, VulkanEnumToStringHeaderGeneratorOptions

from vulkan_object_info_table_base2_header_generator import VulkanObjectInfoTableBase2HeaderGenerator, VulkanObjectInfoTableBase2HeaderGeneratorOptions
from vulkan_state_table_header_generator import VulkanStateTableHeaderGenerator, VulkanStateTableHeaderGeneratorOptions

from vulkan_enum_to_json_body_generator import VulkanEnumToJsonBodyGenerator, VulkanEnumToJsonBodyGeneratorOptions
from vulkan_enum_to_json_header_generator import VulkanEnumToJsonHeaderGenerator, VulkanEnumToJsonHeaderGeneratorOptions
from vulkan_struct_to_json_header_generator import VulkanStructToJsonHeaderGenerator, VulkanStructToJsonHeaderGeneratorOptions
from vulkan_struct_to_json_body_generator import VulkanStructToJsonBodyGenerator, VulkanStructToJsonBodyGeneratorOptions

# Constants
from vulkan_constant_maps_generator import VulkanConstantMapsGenerator, VulkanConstantMapsGeneratorOptions
from vulkan_stype_util_generator import VulkanSTypeUtilGenerator, VulkanSTypeUtilGeneratorOptions

# ToCpp
from vulkan_cpp_struct_generator import VulkanCppStructGenerator, VulkanCppStructGeneratorOptions
from vulkan_cpp_consumer_extension_generator import VulkanCppConsumerExtensionGenerator, VulkanCppConsumerExtensionGeneratorOptions

# Simple timer functions
start_time = None


def start_timer(timeit):
    global start_time
    if timeit:
        start_time = time.process_time()


def end_timer(timeit, msg):
    global start_time
    if timeit:
        end_time = time.process_time()
        write(msg, end_time - start_time, file=sys.stderr)
        start_time = None


# JSON files for customizing code generation
default_blacklists = 'blacklists.json'
default_platform_types = 'platform_types.json'
default_replay_overrides = 'replay_overrides.json'
default_capture_overrides = 'capture_overrides.json'


def _getExtraVulkanHeaders(extraHeadersDir):
    '''
    Recursively get a list of extra Vulkan headers used in the generated code,
    that are included after vulkan/vulkan.h is included
    '''
    extraVulkanHeaders = []
    for child in os.listdir(extraHeadersDir):
        childPath = os.path.join(extraHeadersDir, child)
        if os.path.isdir(childPath):
            extraVulkanHeaders.extend(_getExtraVulkanHeaders(childPath))
        else:
            extraVulkanHeaders.append(childPath)
    return extraVulkanHeaders


def getExtraVulkanHeaders(extraHeadersDir):
    '''
    Get a list of extra Vulkan headers used in the generated code, that are
    included after vulkan/vulkan.h is included
    '''
    return [
        os.path.relpath(header, extraHeadersDir)
        for header in _getExtraVulkanHeaders(extraHeadersDir)
    ]


def extend_xml(dest_tree, src_xml, debug=False):
    '''Extend the parsed vk.xml registry tree with content from another xml file'''
    def merge(dest, src):
        '''Perform a recursive merge of src into dest'''
        leaf_tags = ('command', 'enums', 'extension', 'feature', 'format',
                     'platform', 'spirvcapability', 'spirvextension', 'tag', 'type')
        for src_child in src:
            dest_child = dest.find(src_child.tag)
            if (src_child.tag in leaf_tags) or (dest_child is not None):
                # stop descent and copy if the heirarchy diverges or we reach a leaf tag
                dest.append(src_child)
            else:
                merge(dest_child, src_child)
    merge(dest_tree.getroot(), etree.parse(src_xml).getroot())
    if debug:
        dest_tree.write(os.path.splitext(src_xml)[0] + '_merged.xml')


def make_gen_opts(args):
    """Returns a directory of [ generator function, generator options ] indexed
    by specified short names. The generator options incorporate the following
    parameters:

    args is an parsed argument object; see below for the fields that are used.
    """
    global gen_opts
    gen_opts = {}

    # Output target directory
    directory = args.directory

    # JSON configuration files
    blacklists = os.path.join(args.configs, default_blacklists)
    platform_types = os.path.join(args.configs, default_platform_types)
    replay_overrides = os.path.join(args.configs, default_replay_overrides)
    capture_overrides = os.path.join(args.configs, default_capture_overrides)

    # Copyright text prefixing all headers (list of strings).
    prefix_strings = [
        '/*', '** Copyright (c) 2018-2023 Valve Corporation',
        '** Copyright (c) 2018-2023 LunarG, Inc.',
        '** Copyright (c) 2023 Advanced Micro Devices, Inc.', '**',
        '** Permission is hereby granted, free of charge, to any person obtaining a',
        '** copy of this software and associated documentation files (the "Software"),',
        '** to deal in the Software without restriction, including without limitation',
        '** the rights to use, copy, modify, merge, publish, distribute, sublicense,',
        '** and/or sell copies of the Software, and to permit persons to whom the',
        '** Software is furnished to do so, subject to the following conditions:',
        '**',
        '** The above copyright notice and this permission notice shall be included in',
        '** all copies or substantial portions of the Software.', '**',
        '** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR',
        '** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,',
        '** FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE',
        '** AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER',
        '** LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING',
        '** FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER',
        '** DEALINGS IN THE SOFTWARE.', '*/', ''
    ]

    # Text specific to Vulkan headers
    vk_prefix_strings = [
        '/*',
        '** This file is generated from the Khronos Vulkan XML API Registry.',
        '**', '*/', ''
    ]

    extraVulkanHeaders = []
    if args.headers_dir is not None:
        extraVulkanHeaders = getExtraVulkanHeaders(args.headers_dir)

    #
    # API call decoder generators
    gen_opts['generated_vulkan_decoder.cpp'] = [
        VulkanDecoderBodyGenerator,
        VulkanDecoderBodyGeneratorOptions(
            filename='generated_vulkan_decoder.cpp',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_decoder.h'] = [
        VulkanDecoderHeaderGenerator,
        VulkanDecoderHeaderGeneratorOptions(
            filename='generated_vulkan_decoder.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    #
    # Struct decoder generators
    gen_opts['generated_vulkan_struct_decoders.cpp'] = [
        VulkanStructDecodersBodyGenerator,
        VulkanStructDecodersBodyGeneratorOptions(
            filename='generated_vulkan_struct_decoders.cpp',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_struct_decoders_forward.h'] = [
        VulkanStructDecodersForwardGenerator,
        VulkanStructDecodersForwardGeneratorOptions(
            filename='generated_vulkan_struct_decoders_forward.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_struct_decoders.h'] = [
        VulkanStructDecodersHeaderGenerator,
        VulkanStructDecodersHeaderGeneratorOptions(
            filename='generated_vulkan_struct_decoders.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_decode_pnext_struct.cpp'] = [
        DecodePNextStructGenerator,
        DecodePNextStructGeneratorOptions(
            filename='generated_decode_pnext_struct.cpp',
            directory=directory,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    #
    # Consumer generation
    gen_opts['generated_vulkan_consumer.h'] = [
        VulkanConsumerHeaderGenerator,
        VulkanConsumerHeaderGeneratorOptions(
            class_name='VulkanConsumer',
            base_class_header='vulkan_consumer_base.h',
            is_override=False,
            filename='generated_vulkan_consumer.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_cpp_consumer.h'] = [
        VulkanConsumerHeaderGenerator,
        VulkanConsumerHeaderGeneratorOptions(
        class_name         = 'VulkanCppConsumer',
        base_class_header   = 'vulkan_cpp_consumer_base.h',
        is_override        = True,
        filename          = 'generated_vulkan_cpp_consumer.h',
        directory         = directory,
        blacklists        = blacklists,
        platform_types     = platform_types,
        prefix_text        = prefix_strings + vk_prefix_strings,
        protect_file       = True,
        protect_feature    = False)
    ]

    gen_opts['generated_vulkan_referenced_resource_consumer.h'] = [
        VulkanReferencedResourceHeaderGenerator,
        VulkanReferencedResourceHeaderGeneratorOptions(
            filename='generated_vulkan_referenced_resource_consumer.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_replay_consumer.h'] = [
        VulkanConsumerHeaderGenerator,
        VulkanConsumerHeaderGeneratorOptions(
            class_name='VulkanReplayConsumer',
            base_class_header='vulkan_replay_consumer_base.h',
            is_override=True,
            constructor_args=
            'std::shared_ptr<application::Application> application, const VulkanReplayOptions& options',
            filename='generated_vulkan_replay_consumer.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_cpp_consumer.h'] = [
        VulkanCppConsumerHeaderGenerator,
        VulkanCppConsumerHeaderGeneratorOptions(
        class_name         = 'VulkanCppConsumer',
        base_class_header  = 'vulkan_cpp_consumer_base.h',
        is_override        = True,
        filename           = 'generated_vulkan_cpp_consumer.h',
        directory          = directory,
        blacklists         = blacklists,
        platform_types     = platform_types,
        prefix_text        = prefix_strings + vk_prefix_strings,
        protect_file       = True,
        protect_feature    = False)
    ]

    gen_opts['generated_vulkan_cpp_consumer.cpp'] = [
        VulkanCppConsumerBodyGenerator,
        VulkanCppConsumerBodyGeneratorOptions(
            filename           = 'generated_vulkan_cpp_consumer.cpp',
            directory          = directory,
            blacklists         = blacklists,
            platform_types     = platform_types,
            prefix_text        = prefix_strings + vk_prefix_strings,
            protect_file       = False,
            protect_feature    = False,
            extraVulkanHeaders=extraVulkanHeaders)
    ]

    gen_opts['generated_vulkan_cpp_structs.h'] = [
        VulkanCppStructGenerator,
        VulkanCppStructGeneratorOptions(
            filename           = 'generated_vulkan_cpp_structs.h',
            directory          = directory,
            blacklists         = blacklists,
            platform_types     = platform_types,
            prefix_text        = prefix_strings + vk_prefix_strings,
            protect_file       = True,
            protect_feature    = False)
    ]

    gen_opts['generated_vulkan_cpp_structs.cpp'] = [
        VulkanCppStructGenerator,
        VulkanCppStructGeneratorOptions(
            filename          = 'generated_vulkan_cpp_structs.cpp',
            directory         = directory,
            blacklists        = blacklists,
            platform_types     = platform_types,
            prefix_text        = prefix_strings + vk_prefix_strings,
            protect_file       = False,
            protect_feature    = False)
    ]

    gen_opts['generated_vulkan_cpp_consumer_extension.h'] = [
        VulkanCppConsumerExtensionGenerator,
        VulkanCppConsumerExtensionGeneratorOptions(
            filename          = 'generated_vulkan_cpp_consumer_extension.h',
            directory         = directory,
            blacklists        = blacklists,
            platform_types     = platform_types,
            prefix_text        = prefix_strings + vk_prefix_strings,
            protect_file       = True,
            protect_feature    = False)
    ]

    gen_opts['generated_vulkan_cpp_consumer_extension.cpp'] = [
        VulkanCppConsumerExtensionGenerator,
        VulkanCppConsumerExtensionGeneratorOptions(
            filename          = 'generated_vulkan_cpp_consumer_extension.cpp',
            directory         = directory,
            blacklists        = blacklists,
            platform_types     = platform_types,
            prefix_text        = prefix_strings + vk_prefix_strings,
            protect_file       = False,
            protect_feature    = False)
    ]

    gen_opts['generated_vulkan_replay_consumer.cpp'] = [
        VulkanReplayConsumerBodyGenerator,
        VulkanReplayConsumerBodyGeneratorOptions(
            filename='generated_vulkan_replay_consumer.cpp',
            directory=directory,
            blacklists=blacklists,
            replay_overrides=replay_overrides,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_referenced_resource_consumer.cpp'] = [
        VulkanReferencedResourceBodyGenerator,
        VulkanReferencedResourceBodyGeneratorOptions(
            filename='generated_vulkan_referenced_resource_consumer.cpp',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_struct_handle_mappers.h'] = [
        VulkanStructHandleMappersHeaderGenerator,
        VulkanStructHandleMappersHeaderGeneratorOptions(
            filename='generated_vulkan_struct_handle_mappers.h',
            directory=directory,
            blacklists=blacklists,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_struct_handle_mappers.cpp'] = [
        VulkanStructHandleMappersBodyGenerator,
        VulkanStructHandleMappersBodyGeneratorOptions(
            filename='generated_vulkan_struct_handle_mappers.cpp',
            directory=directory,
            blacklists=blacklists,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_feature_util.cpp'] = [
        VulkanFeatureUtilBodyGenerator,
        VulkanFeatureUtilBodyGeneratorOptions(
            filename='generated_vulkan_feature_util.cpp',
            directory=directory,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    #
    # API call encoder generators
    gen_opts['generated_vulkan_api_call_encoders.h'] = [
        VulkanApiCallEncodersHeaderGenerator,
        VulkanApiCallEncodersHeaderGeneratorOptions(
            filename='generated_vulkan_api_call_encoders.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_api_call_encoders.cpp'] = [
        VulkanApiCallEncodersBodyGenerator,
        VulkanApiCallEncodersBodyGeneratorOptions(
            filename='generated_vulkan_api_call_encoders.cpp',
            directory=directory,
            blacklists=blacklists,
            capture_overrides=capture_overrides,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_command_buffer_util.h'] = [
        VulkanCommandBufferUtilHeaderGenerator,
        VulkanCommandBufferUtilHeaderGeneratorOptions(
            filename='generated_vulkan_command_buffer_util.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_command_buffer_util.cpp'] = [
        VulkanCommandBufferUtilBodyGenerator,
        VulkanCommandBufferUtilBodyGeneratorOptions(
            filename='generated_vulkan_command_buffer_util.cpp',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_dispatch_table.h'] = [
        VulkanDispatchTableGenerator,
        VulkanDispatchTableGeneratorOptions(
            filename='generated_vulkan_dispatch_table.h',
            directory=directory,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_layer_func_table.h'] = [
        LayerFuncTableGenerator,
        LayerFuncTableGeneratorOptions(
            filename='generated_layer_func_table.h',
            directory=directory,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    #
    # Struct encoder generators
    gen_opts['generated_vulkan_struct_encoders.cpp'] = [
        VulkanStructEncodersBodyGenerator,
        VulkanStructEncodersBodyGeneratorOptions(
            filename='generated_vulkan_struct_encoders.cpp',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_struct_encoders.h'] = [
        VulkanStructEncodersHeaderGenerator,
        VulkanStructEncodersHeaderGeneratorOptions(
            filename='generated_vulkan_struct_encoders.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_encode_pnext_struct.cpp'] = [
        EncodePNextStructGenerator,
        EncodePNextStructGeneratorOptions(
            filename='generated_encode_pnext_struct.cpp',
            directory=directory,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_struct_handle_wrappers.h'] = [
        VulkanStructHandleWrappersHeaderGenerator,
        VulkanStructHandleWrappersHeaderGeneratorOptions(
            filename='generated_vulkan_struct_handle_wrappers.h',
            directory=directory,
            blacklists=blacklists,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_struct_handle_wrappers.cpp'] = [
        VulkanStructHandleWrappersBodyGenerator,
        VulkanStructHandleWrappersBodyGeneratorOptions(
            filename='generated_vulkan_struct_handle_wrappers.cpp',
            directory=directory,
            blacklists=blacklists,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_struct_trackers.h'] = [
        VulkanStructTrackersHeaderGenerator,
        VulkanStructTrackersHeaderGeneratorOptions(
            filename='generated_vulkan_struct_trackers.h',
            directory=directory,
            blacklists=blacklists,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_struct_trackers.cpp'] = [
        VulkanStructTrackersBodyGenerator,
        VulkanStructTrackersBodyGeneratorOptions(
            filename='generated_vulkan_struct_trackers.cpp',
            directory=directory,
            blacklists=blacklists,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    #
    # To string generators
    gen_opts['generated_vulkan_enum_to_string.h'] = [
        VulkanEnumToStringHeaderGenerator,
        VulkanEnumToStringHeaderGeneratorOptions(
            filename='generated_vulkan_enum_to_string.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_enum_to_string.cpp'] = [
        VulkanEnumToStringBodyGenerator,
        VulkanEnumToStringBodyGeneratorOptions(
            filename='generated_vulkan_enum_to_string.cpp',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_object_info_table_base2.h'] = [
        VulkanObjectInfoTableBase2HeaderGenerator,
        VulkanObjectInfoTableBase2HeaderGeneratorOptions(
            filename='generated_vulkan_object_info_table_base2.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_state_table.h'] = [
        VulkanStateTableHeaderGenerator,
        VulkanStateTableHeaderGeneratorOptions(
            filename='generated_vulkan_state_table.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_json_consumer.h'] = [
        VulkanExportJsonConsumerHeaderGenerator,
        VulkanExportJsonConsumerHeaderGeneratorOptions(
            class_name='VulkanExportJsonConsumer',
            base_class_header='vulkan_json_consumer_base.h',
            is_override=True,
            filename='generated_vulkan_json_consumer.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_json_consumer.cpp'] = [
        VulkanExportJsonConsumerBodyGenerator,
        VulkanExportJsonConsumerBodyGeneratorOptions(
            filename='generated_vulkan_json_consumer.cpp',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_struct_to_json.h'] = [
        VulkanStructToJsonHeaderGenerator,
        VulkanStructToJsonHeaderGeneratorOptions(
            filename='generated_vulkan_struct_to_json.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_struct_to_json.cpp'] = [
        VulkanStructToJsonBodyGenerator,
        VulkanStructToJsonBodyGeneratorOptions(
            filename='generated_vulkan_struct_to_json.cpp',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=False,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_enum_to_json.h'] = [
        VulkanEnumToJsonHeaderGenerator,
        VulkanEnumToJsonHeaderGeneratorOptions(
            filename='generated_vulkan_enum_to_json.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefixText=prefix_strings + vk_prefix_strings,
            protectFile=True,
            protectFeature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_enum_to_json.cpp'] = [
        VulkanEnumToJsonBodyGenerator,
        VulkanEnumToJsonBodyGeneratorOptions(
            filename='generated_vulkan_enum_to_json.cpp',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefixText=prefix_strings + vk_prefix_strings,
            protectFile=False,
            protectFeature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_constant_maps.h'] = [
        VulkanConstantMapsGenerator,
        VulkanConstantMapsGeneratorOptions(
            filename='generated_vulkan_constant_maps.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

    gen_opts['generated_vulkan_stype_util.h'] = [
        VulkanSTypeUtilGenerator,
        VulkanSTypeUtilGeneratorOptions(
            filename='generated_vulkan_stype_util.h',
            directory=directory,
            blacklists=blacklists,
            platform_types=platform_types,
            prefix_text=prefix_strings + vk_prefix_strings,
            protect_file=True,
            protect_feature=False,
            extraVulkanHeaders=extraVulkanHeaders
        )
    ]

def gen_target(args):
    """Generate a target based on the options in the matching gen_opts{} object.
    This is encapsulated in a function so it can be profiled and/or timed.
    The args parameter is an parsed argument object containing the following
    fields that are used:
      target - target to generate
      directory - directory to generate it in
      protect - True if re-inclusion wrappers should be created
      extensions - list of additional extensions to include in generated
      interfaces
    """
    # Create generator options with specified parameters
    make_gen_opts(args)

    if args.target in gen_opts:
        create_generator = gen_opts[args.target][0]
        options = gen_opts[args.target][1]

        if not args.quiet:
            write('* Building', options.filename, file=sys.stderr)
            write(
                '* options.versions          =',
                options.versions,
                file=sys.stderr
            )
            write(
                '* options.emitversions      =',
                options.emitversions,
                file=sys.stderr
            )
            write(
                '* options.defaultExtensions =',
                options.defaultExtensions,
                file=sys.stderr
            )
            write(
                '* options.addExtensions     =',
                options.addExtensions,
                file=sys.stderr
            )
            write(
                '* options.removeExtensions  =',
                options.removeExtensions,
                file=sys.stderr
            )
            write(
                '* options.emitEtensions    =',
                options.emitExtensions,
                file=sys.stderr
            )

        gen = create_generator(
            err_file=err_warn, warn_file=err_warn, diag_file=diag
        )

        return (gen, options)
    else:
        write(
            'No generator options for unknown target:',
            args.target,
            file=sys.stderr
        )
        return None


# -feature name
# -extension name
# For both, "name" may be a single name, or a space-separated list
# of names, or a regular expression.
if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument('-debug', action='store_true', help='Enable debugging')
    parser.add_argument(
        '-dump', action='store_true', help='Enable dump to stderr'
    )
    parser.add_argument(
        '-diagfile',
        action='store',
        default=None,
        help='Write diagnostics to specified file'
    )
    parser.add_argument(
        '-errfile',
        action='store',
        default=None,
        help='Write errors and warnings to specified file instead of stderr'
    )
    parser.add_argument(
        '-noprotect',
        dest='protect',
        action='store_false',
        help='Disable inclusion protection in output headers'
    )
    parser.add_argument(
        '-profile', action='store_true', help='Enable profiling'
    )
    parser.add_argument(
        '-registry',
        action='store',
        default='vk.xml',
        help='Use specified registry file instead of vk.xml'
    )
    parser.add_argument(
        '-video',
        action='store',
        default='video.xml',
        help='Use specified video file instead of video.xml'
    )
    parser.add_argument(
        '-headers-dir',
        dest='headers_dir',
        action='store',
        default=None,
        help='\n'.join(
            [
                'Path to a directory that holds additional Vulkan header files required to build.',
                'These header files are included directly after the Vulkan header in all generated files.'
            ]
        )
    )
    parser.add_argument('-time', action='store_true', help='Enable timing')
    parser.add_argument(
        '-validate', action='store_true', help='Enable group validation'
    )
    parser.add_argument(
        '-o',
        action='store',
        dest='directory',
        default='.',
        help='Create target and related files in specified directory'
    )
    parser.add_argument(
        'target', metavar='target', nargs='?', help='Specify target'
    )
    parser.add_argument(
        '-quiet',
        action='store_true',
        default=True,
        help='Suppress script output during normal execution.'
    )
    parser.add_argument(
        '-verbose',
        action='store_false',
        dest='quiet',
        default=True,
        help='Enable script output during normal execution.'
    )
    parser.add_argument(
        '-configs',
        action='store',
        dest='configs',
        default='.',
        help=
        'Specify directory containing JSON configuration files for generators'
    )

    args = parser.parse_args()

    # create error/warning & diagnostic files
    if (args.errfile):
        err_warn = open(args.errfile, 'w', encoding='utf-8')
    else:
        err_warn = sys.stderr

    if (args.diagfile):
        diag = open(args.diagfile, 'w', encoding='utf-8')
    else:
        diag = None

    (gen, options) = gen_target(args)

    reg = Registry(gen, options)

    ## @note We parse vk.xml to an in-memory element tree and then extract the info we need
    ## from that into the Registry object once per output file we generate rather than once
    ## per run of the top-level generation script.
    start_timer(args.time)
    tree = etree.parse(args.registry)
    gen.VIDEO_TREE = etree.parse(args.video)
    end_timer(args.time, '* Time to make ElementTree =')

    # Apply any temporary patches to the xml to allow correct code generation
    # from it. These should be reviewed and removed when the xml is fixed upstream.
    start_timer(args.time)
    # There are no current patches needed, but here is an example of how to patch the XML
    # after parsing to add a missing single attribute:
    # Workaround 1.3.264 VkFrameBoundaryEXT.pTag lacking len field:
    # <https://github.com/KhronosGroup/Vulkan-Docs/pull/2240>
    # if ptag_member := tree.find('types/type[@name="VkFrameBoundaryEXT"]/member[name="pTag"]'):
    #    ptag_member.set('len', 'tagSize')
    end_timer(args.time, '* Time to patch ElementTree =')

    # Extend the vk.xml tree with XML files from the config dir
    for filename in os.listdir(args.configs):
        if filename.endswith('.xml'):
            extend_xml(tree, os.path.join(args.configs, filename))

    start_timer(args.time)
    reg.loadElementTree(tree)
    end_timer(args.time, '* Time to parse ElementTree =')

    if (args.validate):
        reg.validateGroups()

    if (args.dump):
        write('* Dumping registry to regdump.txt', file=sys.stderr)
        reg.dumpReg(filehandle=open('regdump.txt', 'w', encoding='utf-8'))

    if (args.debug):
        pdb.run('reg.api_gen()')
    else:
        start_timer(args.time)
        reg.apiGen()
        end_timer(args.time, '* Time to generate ' + options.filename + ' =')

    if not args.quiet:
        write('* Generated', options.filename, file=sys.stderr)



#!/usr/bin/env python3
#
# Copyright (c) 2021 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from base_generator import BaseGenerator, write
from dx12_base_generator import Dx12BaseGenerator, Dx12GeneratorOptions


class Dx12ConsumerHeaderGeneratorOptions(Dx12GeneratorOptions):
    """Options for generating a C++ class for Dx12 capture file replay."""

    def __init__(
        self,
        constructor_args='',
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True
    ):
        Dx12GeneratorOptions.__init__(
            self, blacklists, platform_types, filename, directory, prefix_text,
            protect_file, protect_feature
        )
        self.constructor_args = constructor_args


class Dx12ConsumerHeaderGenerator(Dx12BaseGenerator):
    """Generates C++ functions responsible for consuming Dx12 API calls."""
    constructor_args = ''

    def __init__(
        self,
        source_dict,
        dx12_prefix_strings,
        err_file=sys.stderr,
        warn_file=sys.stderr,
        diag_file=sys.stdout
    ):
        Dx12BaseGenerator.__init__(
            self, source_dict, dx12_prefix_strings, err_file, warn_file,
            diag_file
        )

    def beginFile(self, gen_opts):
        """Method override."""
        BaseGenerator.beginFile(self, gen_opts)

        if gen_opts.constructor_args:
            self.constructor_args = gen_opts.constructor_args

        self.write_include()
        write('GFXRECON_BEGIN_NAMESPACE(gfxrecon)', file=self.outFile)
        write('GFXRECON_BEGIN_NAMESPACE(decode)', file=self.outFile)
        self.newline()

    def generate_feature(self):
        """Method override."""
        Dx12BaseGenerator.generate_feature(self)
        self.write_dx12_consumer_class('')

    def write_include(self):
        code = (
            "\n"
            "#include \"decode/dx12_consumer_base.h\"\n"
            "#include \"generated_dx12_struct_decoders.h\"\n"
            "#include \"decode/custom_dx12_struct_decoders.h\"\n"
            "\n"
        )
        write(code, file=self.outFile)

    def get_consumer_function_body(self, class_name, method_info, return_type, return_value):
        return '{}'

    def change_param_type(self, param):
        value = self.get_value_info(param)
        type_name = self.make_decoded_param_type(value)
        if value.pointer_count > 0 and type_name.find('Decoder') != -1:
            type_name += '*'
        return type_name

    def get_consumer_function(
        self, class_name, method_info, consumer_type, indent, function_class
    ):
        parameters = '        const ApiCallInfo& call_info,\n'
        class_method_name = method_info['name']
        if class_name:
            parameters += '        format::HandleId object_id'
            class_method_name = class_name + '_' + class_method_name

        rtn_type = method_info['rtnType']
        return_value = None
        if rtn_type.find('void ') == -1 or rtn_type.find('void *') != -1:
            rtn_type1 = self.clean_type_define(rtn_type)
            return_value = self.get_return_value_info(
                rtn_type1, class_method_name
            )
            rtn_type1 = self.make_decoded_param_type(return_value)
            if rtn_type1.find('Decoder') != -1:
                rtn_type1 += '*'

            if class_name:
                parameters += ',\n'
            parameters += '        ' + rtn_type1 + ' ' + return_value.name

        space_index = 0
        for p in method_info['parameters']:
            if parameters:
                parameters += ',\n'
            parameters += '        '
            parameters += self.change_param_type(p)
            parameters += ' '
            parameters += p['name']

            if 'array_size' in p:
                array_length = p['array_size']
                parameters += ' '

                if 'multi_dimensional_array' in p:
                    p['multi_dimensional_array']

                    if 'multi_dimensional_array_size' in p:
                        multi_dimensional_array_size = p[
                            'multi_dimensional_array_size']

                        array_sizes = multi_dimensional_array_size.split("x")
                        for size in array_sizes:
                            parameters += '['
                            parameters += size
                            parameters += ']'
                else:
                    parameters += '['
                    parameters += array_length
                    parameters += ']'

            while True:
                space_index = parameters.find(' ', space_index) + 1

                if space_index != 0 and (
                    parameters[space_index] == '*'
                    or parameters[space_index - 2] == '('
                    or parameters[space_index] == '('
                    or parameters[space_index] == ')'
                ):
                    parameters = parameters[:space_index
                                            - 1] + parameters[space_index:]
                elif space_index == 0:
                    break

        _class_name = ''
        if class_name:
            _class_name = '_' + class_name

        code = '{}void {}Process{}_{}(\n'\
            '{}){}\n\n'.format(
                indent,
                function_class,
                _class_name,
                method_info['name'],
                parameters,
                self.get_consumer_function_body(class_name, method_info, rtn_type, return_value))
        return code

    def write_constructor_class(self, consumer_type):
        if self.constructor_args:
            arg_list = ', '.join(
                [
                    arg.split(' ')[-1]
                    for arg in self.constructor_args.split(',')
                ]
            )
            return 'class Dx12{0}Consumer : public Dx12{0}ConsumerBase\n'\
                      '{{\n'\
                      '  public:\n'\
                      '    Dx12{0}Consumer({1}) : Dx12{0}ConsumerBase({2}) {{}}\n'\
                      '    virtual ~Dx12{0}Consumer() override {{}}\n'.format(
                          consumer_type, self.constructor_args, arg_list)
        else:
            return 'class Dx12{0}Consumer : public Dx12{0}ConsumerBase\n'\
                      '{{\n'\
                      '  public:\n'\
                      '    Dx12{0}Consumer(){{}}\n'\
                      '    virtual ~Dx12{0}Consumer() override {{}}\n'.format(
                          consumer_type)

    def get_decoder_class_define(self, consumer_type):
        declaration = self.write_constructor_class(consumer_type)

        indent = '    virtual '
        function_class = ''
        class_end = '};\n'
        return (declaration, indent, function_class, class_end)

    def write_dx12_consumer_class(self, consumer_type):
        declaration, indent, function_class, class_end = self.get_decoder_class_define(
            consumer_type
        )
        code = declaration

        header_dict = self.source_dict['header_dict']
        for k, v in header_dict.items():
            code_length = len(code)
            for m in v.functions:
                if self.is_required_function_data(m) and (
                    not self.is_cmd_black_listed(m['name'])
                ):
                    code += self.get_consumer_function(
                        '', m, consumer_type, indent, function_class
                    )

            for class_name, class_value in v.classes.items():
                if self.is_required_class_data(class_value):
                    for m in class_value['methods']['public']:
                        if not self.is_method_black_listed(
                            class_name, m['name']
                        ):
                            code += self.get_consumer_function(
                                class_name, m, consumer_type, indent,
                                function_class
                            )

            code_length2 = len(code)
            if code_length2 > code_length:
                code = code[:code_length] + self.dx12_prefix_strings.format(
                    k
                ) + '\n' + code[code_length:]

        code += class_end
        write(code, file=self.outFile)

    def endFile(self):
        """Method override."""
        self.newline()
        write('GFXRECON_END_NAMESPACE(decode)', file=self.outFile)
        write('GFXRECON_END_NAMESPACE(gfxrecon)', file=self.outFile)

        # Finish processing in superclass
        BaseGenerator.endFile(self)



#!/usr/bin/env python3
#
# Copyright (c) 2023 LunarG, Inc.
# Copyright (c) 2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from base_generator import BaseGenerator, write
from dx12_base_generator import Dx12BaseGenerator


class Dx12CallIdToStringHeaderGenerator(Dx12BaseGenerator):

    """Generates C++ function responsible for converting Dx12 ApiCallId to string."""

    def __init__(
        self,
        source_dict,
        dx12_prefix_strings,
        err_file=sys.stderr,
        warn_file=sys.stderr,
        diag_file=sys.stdout
    ):
        Dx12BaseGenerator.__init__(
            self, source_dict, dx12_prefix_strings, err_file, warn_file,
            diag_file
        )

    def beginFile(self, gen_opts):
        BaseGenerator.beginFile(self, gen_opts)

        self.write_include()
        write('GFXRECON_BEGIN_NAMESPACE(gfxrecon)', file=self.outFile)
        write('GFXRECON_BEGIN_NAMESPACE(util)', file=self.outFile)
        self.newline()

    def generate_feature(self):
        Dx12BaseGenerator.generate_feature(self)
        self.write_function_call()

    def write_include(self):
        code = ("\n"
                "#include \"format/api_call_id.h\"\n"
                "#include <string>\n"
               "\n")
        write(code, file=self.outFile)

    def write_function_call(self):
        code = (
            "inline std::wstring GetDx12CallIdString(format::ApiCallId call_id){}\n"
            "\n"
            .format(
                self.get_function_call_body()
            )
        )
        write(code, file=self.outFile)

    def get_function_call_body(self):
        code = '\n'\
               '{\n'\
               '    std::wstring out = L"Unknown_ApiCallId";\n'\
               '    switch (call_id)\n'\
               '    {\n'

        header_dict = self.source_dict['header_dict']
        for k, v in header_dict.items():
            for m in v.functions:
                if self.is_required_function_data(m):
                    code += (
                        "    case format::ApiCallId::ApiCall_{0}:\n"
                        "        out = L\"{0}\";\n"
                        "        break;\n".format(m['name'])
                    )

            for class_name, class_value in v.classes.items():
                if self.is_required_class_data(class_value):
                    for m in class_value['methods']['public']:
                        code += (
                            "    case format::ApiCallId::ApiCall_{0}_{1}:\n"
                            "        out = L\"{0}_{1}\";\n"
                            "        break;\n".format(
                                class_name, m['name']
                            )
                        )


        code += '    default:\n'\
                '        break;\n'\
                '    }\n'\
                '    return out;\n'\
                '}'

        return code


    def endFile(self):
        write('GFXRECON_END_NAMESPACE(util)', file=self.outFile)
        write('GFXRECON_END_NAMESPACE(gfxrecon)', file=self.outFile)

        # Finish processing in superclass
        BaseGenerator.endFile(self)



#!/usr/bin/env python3
#
# Copyright (c) 2023 LunarG, Inc.
# Copyright (c) 2023 Valve Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
#
# Stub for redirecting to new gfxrecon-capture-vulkan command
# Usage:
#
#     gfxrecon-capture.py [args]
#
#         args is a command-specific argument list

import argparse
import os
import sys
import subprocess
import platform

if sys.version_info < (3, 7):
    print("Error: Python 3.7 or later is required; you're using:", sys.version)
    sys.exit(1)

argv = sys.argv
argc = len(sys.argv)

if __name__ == '__main__':

    # We don't support running under Cygwin Python
    if sys.platform == 'cygwin':
        print(os.path.basename(__file__) + ' error: Cygwin Python not supported')
        sys.exit(1)

    print("Warning: the 'gfxrecon-capture.py' script is deprecated and will be removed in a future update. Use 'gfxrecon-capture-vulkan.py' instead.", flush=True)
    result = subprocess.run(["gfxrecon-capture-vulkan.py",] + argv[1:], shell = True)
    sys.exit(result.returncode)



#!/usr/bin/python3
#
# Copyright (c) 2024 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from base_generator import BaseGenerator, BaseGeneratorOptions, write
from reformat_code import format_cpp_code

VulkanSTypeUtilGeneratorOptions = BaseGeneratorOptions

class VulkanSTypeUtilGenerator(BaseGenerator):
    """VulkanSTypeUtilGenerator - subclass of BaseGenerator.
    Generates C++ utility header to do compile-time lookups between Vulkan
    structure types and their corresponding VkStructureType values
    """
    def __init__(self, **kwargs):
        BaseGenerator.__init__(
            self,
            process_cmds=False,
            process_structs=False,
            feature_break=False,
            **kwargs
        )
        # These are dummy structures that do not have their own VkStructureType values
        self.missing_stypes = ['VkBaseInStructure', 'VkBaseOutStructure']

    def beginFile(self, gen_opts):
        BaseGenerator.beginFile(self, gen_opts)
        write('#include "format/platform_types.h"', file=self.outFile)
        write('#include "util/defines.h"', file=self.outFile)
        self.newline()
        self.includeVulkanHeaders(gen_opts)
        self.newline()
        write('GFXRECON_BEGIN_NAMESPACE(gfxrecon)', file=self.outFile)
        write('GFXRECON_BEGIN_NAMESPACE(util)', file=self.outFile)
        self.newline()
        write(format_cpp_code('''
            // Instantiating the primary template indicates that either the template was
            // called with an invalid Vulkan struct type or that the code generation is out
            // of date, both of which are errors
            template <typename T> VkStructureType GetSType() = delete;'''), file=self.outFile)
        self.newline()

    def genStruct(self, typeinfo, typename, alias):
        if not alias and typename not in self.missing_stypes:
            stype = self.make_structure_type_enum(typeinfo, typename)
            if stype:
                write(f'template <> constexpr VkStructureType GetSType<{typename}>(){{ return {stype}; }}', file=self.outFile)

    def endFile(self):
        write('GFXRECON_END_NAMESPACE(util)', file=self.outFile)
        write('GFXRECON_END_NAMESPACE(gfxrecon)', file=self.outFile)

        # Finish processing in superclass
        BaseGenerator.endFile(self)
